"""Base class for all motor based devices
"""

import logging

from beamline_console.constants import APP_NAME

from beamline_console.devices.tangomotor import TangoMotor
from beamline_console.devices.dummymotor import DummyMotor

logger = logging.getLogger(APP_NAME)


# ----------------------------------------------------------------------
class MotorBasedDevice:
    """
    """

    # ----------------------------------------------------------------------
    def __init__(self, device_name, beamline_hal):
        """
        """
        self.name = device_name

        self._history = {}
        self.motors = {}

        self.beamline_hal = beamline_hal

        logger.debug(f"Instantiating device object {device_name}")

    # ----------------------------------------------------------------------
    def motor(self, motor_name):
        """
        Returns:
            motor handle
        """
        if motor_name in self.motors:
            return self.motors[motor_name]

        return None

    # ----------------------------------------------------------------------
    def progress(self):
        progress = []
        for m in list(self.motors.values()):
            if m.state() == "moving":
                progress.append(m.progress())
        if progress:
            return min(progress)
        else:
            return None

    # ----------------------------------------------------------------------
    def state(self):
        """
        """
        if any([m.state() == "moving" for m in list(self.motors.values())]):
            return "moving"
        
        if all([m.state() == "idle" for m in list(self.motors.values())]):
            return "ready"          # or "idle"
        
        if any([m.state() == "fault" for m in list(self.motors.values())]):
            return "fault"
        
        return "unknown"
        #return "ready"

    # ----------------------------------------------------------------------
    def move(self, motor_name, target_position, mode="async"):
        """Move given motor to the target position.
        """
        motor = self.motor(motor_name)
        if motor:
            motor.move_absolute(target_position, mode)
        else:
            raise RuntimeError(f"Invalid motor name {motor_name}")

    # ----------------------------------------------------------------------
    def stop_motors(self):
        """Stop all motors.
        """
        logger.info(f"Stopping device {self.name}")
        for motor in list(self.motors.values()):
            motor.stop()

    # ----------------------------------------------------------------------
    def initialize(self):
        """
        """
        status = True

        for motor in list(self.motors.values()):
            status = status*motor.initialize()

        return status

    # ----------------------------------------------------------------------
    def access_level_changed(self, access_level):
        for motor in list(self.motors.values()):
            motor.access_level_changed(access_level)

    # ----------------------------------------------------------------------
    def from_xml(self, xml_node):
        """Load device's configuration from a given XML node.
        """
        general_access = xml_node.getAttribute('access')
        if not general_access:
            general_access = "user"

        for motor_node in xml_node.getElementsByTagName("motor"):
            name = str(motor_node.getAttribute("name"))
            if name in self.motors:
                raise RuntimeError(f'Duplication of motors name {name} in {self.name}')
            motor = self._make_motor(motor_node, general_access)
            if motor:
                self.motors[name] = motor

    # ----------------------------------------------------------------------
    def _make_motor(self, motor_node, general_access):
        """Factory method
        """

        motor_type = motor_node.getAttribute("type")
        if  motor_type== "tango":
            motor = TangoMotor()
            motor.tango_server = str(motor_node.getAttribute("tango_server"))
            motor.updated_properties(motor_node)

        elif motor_type == "dummy":
            motor = DummyMotor()
        else:
            raise ValueError(f"Invalid motor type {motor_type}")

        access = motor_node.getAttribute('access')
        motor.set_access(access if access else general_access)

        motor.name = str(motor_node.getAttribute("name"))

        try:
            motor.user_low_limit = float(motor_node.getAttribute("low_limit_user"))
            motor.user_high_limit = float(motor_node.getAttribute("high_limit_user"))

        except ValueError:
            print(f"(motordevice.py) no user limits for motor {motor.name} defined")

        print(f"(motordevice.py) loaded motor {motor.name}")
    
        return motor

    # ----------------------------------------------------------------------
    def list_motors(self):
        """
        Returns:
            [str] list of motor names
        """
        return list(self.motors.keys())


"""Basic motor's properties show/edit dialog.
"""

import math
import subprocess

from PyQt5 import QtWidgets, QtCore

from beamline_console.logger.logit import Logit
from beamline_console.gui.MotorPropsDialog_ui import Ui_MotorPropsDialog


# ---------------------------------------------------------------------- 
class MotorPropsDialog(QtWidgets.QDialog, Logit):
    """
    """

    VALIDATOR_REFRESH = 500               # ms
    EPSILON = 1e-8

    NORMAL_STYLE = "QLineEdit {background-color: white};"
    ERROR_STYLE = "QLineEdit {background-color: #FFC4C4};"
    CHANGED_STYLE = "QLineEdit {background-color: #FFFCC4};"

    PARAMS_EDITOR = "atkpanel"

    # ---------------------------------------------------------------------- 
    def __init__(self, motor, parent=None):
        """
        Args:
            motor (see beamlinehal/motor)
        """
        QtWidgets.QDialog.__init__(self, parent)
        Logit.__init__(self)

        self._motor = motor             # HAL handle

        self._ui = Ui_MotorPropsDialog()
        self._ui.setupUi(self)
      
        self._ui.btnSaveProps.clicked.connect(self._saveProperties)
        self._ui.btnCalibrate.clicked.connect(self._calibrateMotorPosition)
        self._ui.btnUndo.clicked.connect(self._undoChanges)
        self._ui.btnAllParams.clicked.connect(self._editAllParams)
   
            # backup current values
        self._conversionBak = self._motor.parameter("Conversion")
        self._slewRateBak = self._motor.parameter("SlewRate")
        self._accelerationBak = self._motor.parameter("Acceleration")
        self._backlashBak = self._motor.parameter("StepBacklash")
    
        self._lowLimitBak = self._motor.parameter("UnitLimitMin")
        self._highLimitBak = self._motor.parameter("UnitLimitMax")
        self._decimalsBak = self._motor.decimals


        self._baseRateBak = self._motor.parameter("BaseRate")

            # user limits (on top of TANGO)
        self._lowLimitUserBak = self._motor.user_low_limit
        self._highLimitUserBak = self._motor.user_high_limit

        self._ui.chbUserLimits.toggled.connect(self._enableUserLimits)
        self._ui.chbUserLimits.setChecked(bool(self._motor.user_low_limit and
                                               self._motor.user_high_limit))
        self._refreshGui()

        self._validatorTimer = QtCore.QTimer(self)
        self._validatorTimer.timeout.connect(self._maybeColor)
        self._validatorTimer.start(self.VALIDATOR_REFRESH)
        
            # in case of multiple screens center on the primary one
        frame = self.frameGeometry()

        desktop = QtWidgets.QApplication.desktop()
        idxPrimary = desktop.primaryScreen()
        centerPoint = desktop.screenGeometry(idxPrimary).center()
        
        frame.moveCenter(centerPoint)
        self.move(frame.topLeft())

        #self.setModal(True)

    # ---------------------------------------------------------------------- 
    def _refreshGui(self):
        """
        """
        self._ui.lbMotorName.setText(str(self._motor.name))
        self._ui.lbTangoServer.setText(str(self._motor.tango_server))

        self._ui.leConversion.setText("{:.2f}".format(self._motor.parameter("Conversion")))
        self._ui.leCalibration.setText("0.0")
    
        self._ui.leSlewRate.setText(str(self._motor.parameter("SlewRate")))
        self._ui.leAcceleration.setText(str(self._motor.parameter("Acceleration")))
        self._ui.leBacklash.setText(str(self._motor.parameter("StepBacklash")))
    
        self._ui.leLowLimit.setText("{:.2f}".format(self._motor.parameter("UnitLimitMin")))
        self._ui.leHighLimit.setText("{:.2f}".format(self._motor.parameter("UnitLimitMax")))
        self._ui.leDecimals.setText(f"{self._motor.decimals:d}")
        
        self._ui.leBaseRate.setText("{:.2f}".format(self._motor.parameter("BaseRate")))

        if (self._motor.user_low_limit and
            self._motor.user_high_limit):
            self._ui.leLowLimitUser.setText(f"{self._motor.user_low_limit:.2f}")
            self._ui.leHighLimitUser.setText(f"{self._motor.user_high_limit:.2f}")

        self._ui.frameUserLimits.setEnabled(self._ui.chbUserLimits.isChecked())

    # ---------------------------------------------------------------------- 
    def _saveProperties(self):
        """Validate user input and save changes to the TANGO database.
        """
        try:
            self._motor.set_parameter("Conversion", float(self._ui.leConversion.text()))
            self._motor.set_parameter("SlewRate", int(self._ui.leSlewRate.text()))
            self._motor.set_parameter("Acceleration", int(self._ui.leAcceleration.text()))
            self._motor.set_parameter("StepBacklash", int(self._ui.leBacklash.text()))
    
            self._motor.set_parameter("UnitLimitMin", float(self._ui.leLowLimit.text()))
            self._motor.set_parameter("UnitLimitMax", float(self._ui.leHighLimit.text()))
            self._motor.decimals = int(self._ui.leDecimals.text())

            if self._ui.chbUserLimits.isChecked():
                self._motor.user_low_limit =  float(self._ui.leLowLimitUser.text())
                self._motor.user_high_limit = float(self._ui.leHighLimitUser.text())
            else:
                self._motor.user_low_limit =  None
                self._motor.user_high_limit = None
            
            self._motor.set_parameter("BaseRate", float(self._ui.leBaseRate.text()))

            self.logger.info(f"motor {self._motor.name} properties changed!")

        except ValueError as err:
            self.logger.warning(f"{err}")
            QtWidgets.QMessageBox.warning(self, "Experiment Control",
                                      f"Exception: {str(err)}", QtWidgets.QMessageBox.Ok)
            return
   
        self.properties_changed.emit()
    
        super().accept() 
  
    # ---------------------------------------------------------------------- 
    def _calibrateMotorPosition(self):
        """
        """
        physicalPos = float(self._ui.leCalibration.text())
        self._motor.calibrate(physicalPos)
      
        self.logger.info(f"Motor {self._motor.name} recalibrated, position {physicalPos:.2f}")

    # ----------------------------------------------------------------------
    def _editAllParams(self):
        """
        """
        subprocess.Popen([self.PARAMS_EDITOR, self._motor.tango_server])

    # ---------------------------------------------------------------------- 
    def _enableUserLimits(self, flag):
        """
        """
        self._ui.frameUserLimits.setEnabled(flag)

        if flag:
            self._ui.leLowLimitUser.setText("%.2f" % self._motor.parameter("UnitLimitMin"))
            self._ui.leHighLimitUser.setText("%.2f" % self._motor.parameter("UnitLimitMax"))
        else:
            self._ui.leLowLimitUser.clear()
            self._ui.leHighLimitUser.clear()

    # ---------------------------------------------------------------------- 
    def _maybeColor(self):
        """
        """
        self._colorLineEdit(self._ui.leConversion, self._conversionBak)
        self._colorLineEdit(self._ui.leCalibration, 0.0)
    
        self._colorLineEdit(self._ui.leSlewRate, self._slewRateBak)
        self._colorLineEdit(self._ui.leAcceleration, self._accelerationBak)
        self._colorLineEdit(self._ui.leBacklash, self._backlashBak)
    
        self._colorLineEdit(self._ui.leLowLimit, self._lowLimitBak)
        self._colorLineEdit(self._ui.leHighLimit, self._highLimitBak)

        if self._ui.chbUserLimits.isChecked():
            if not self._lowLimitUserBak:
                self._lowLimitUserBak = self._lowLimitBak
            if not self._highLimitUserBak:
                self._highLimitUserBak = self._highLimitBak

            self._colorLineEdit(self._ui.leLowLimitUser, self._lowLimitUserBak)
            self._colorLineEdit(self._ui.leHighLimitUser, self._highLimitUserBak)

    # ---------------------------------------------------------------------- 
    def _colorLineEdit(self, lineEdit, prevValue):
        """
        """
        try:
            value = float(lineEdit.text())
    
            if math.fabs(value - float(prevValue)) > self.EPSILON:
                lineEdit.setStyleSheet(self.CHANGED_STYLE)
                lineEdit.setToolTip(f"Previous value {prevValue}")
            else:
                lineEdit.setStyleSheet(self.NORMAL_STYLE)
                lineEdit.setToolTip("")

        except ValueError as err:
            lineEdit.setStyleSheet(self.ERROR_STYLE)
            lineEdit.setToolTip("Not a number")

    # ---------------------------------------------------------------------- 
    def _undoChanges(self):
        """
        """
        self._ui.leConversion.setText(f"{self._conversionBak:.2f}")
        self._ui.leCalibration.setText("0.0")

        self._ui.leSlewRate.setText(str(self._slewRateBak))
        self._ui.leAcceleration.setText(str(self._accelerationBak))
        self._ui.leBacklash.setText(str(self._backlashBak))
    
        self._ui.leLowLimit.setText(f"{self._lowLimitBak:.2f}")
        self._ui.leHighLimit.setText(f"{self._highLimitBak:.2f}")

        if self._ui.chbUserLimits.isChecked():
            if self._lowLimitUserBak and self._highLimitUserBak:
                self._ui.leLowLimitUser.setText(f"{self._lowLimitUserBak:.2f}")
                self._ui.leHighLimitUser.setText(f"{self._highLimitUserBak:.2f}")
            else:
                self._ui.leLowLimitUser.clear()
                self._ui.leHighLimitUser.clear()
 



"""Displays currently being selected device(s).
"""

import sys
import json
import time
from PyQt5 import QtWidgets, QtCore

from beamline_console.logger.logit import Logit
from beamline_console.devicecontrol.motoritemwidget import MotorItemWidget

from beamline_console.gui.DeviceWidget_ui import Ui_DeviceWidget

# ----------------------------------------------------------------------
class DeviceWidget(QtWidgets.QWidget, Logit):
    """
    """
    save_snapshot = QtCore.pyqtSignal(str)  # motors selection info

    DATETIME = "%Y/%m/%d %H:%M:%S"

    # ----------------------------------------------------------------------
    def __init__(self, parent):
        """
        Args:
            parent (QWidget)
        """

        QtWidgets.QWidget.__init__(self, parent)
        Logit.__init__(self)

        self._ui = Ui_DeviceWidget()
        self._ui.setupUi(self)

        self._selected_devices = []
        self.device_tuple_map = {}

        self._beamline_hal = None
        self._settings = None

        self._device_widget_layout = QtWidgets.QVBoxLayout(self._ui.wi_device_container)

        self._timer = QtCore.QTimer(self)
        self._timer.timeout.connect(self.refresh_gui)

        self._ui.but_save.clicked.connect(self._save_snapshot)
        self._ui.but_stop.clicked.connect(self.stop_motors)

    # ----------------------------------------------------------------------
    def initialize(self, beamline_hal, settings):
        """
        """
        self._beamline_hal = beamline_hal
        self._settings = settings

        for dev_name, dev_handle in sorted(self._beamline_hal.device_map.items()):
            motors = []
            for motor_handle in list(dev_handle.motors.values()):
                motor = MotorItemWidget(self._beamline_hal, dev_handle, motor_handle, self)
                motors.append(motor)

            self.device_tuple_map[dev_name] = motors

        self._timer.start(int(self._settings.option("beamline_ui", "selected_device_refresh_period")))

        self.logger.info("DeviceWidget initialized successfully")

    # ----------------------------------------------------------------------
    def refresh_gui(self):
        """Synchronize GUI (beamline state can be changed by Jive, Astor, etc.)
        """
        for device_name in self._selected_devices:
            try:
                widgets = self.device_tuple_map[device_name]
                for item in widgets:
                    item.refresh_gui()
            except KeyError:
                pass

    # ----------------------------------------------------------------------
    def stop_motors(self):
        """
        """
        try:
            self._beamline_hal.group_action("stop_motors")
            time.sleep(0.5)

            self.sync_gui("absolute")

        except Exception as err:
            self.logger.error(err, exc_info=sys.exc_info())
            raise

            # non motor devs TODO

        self.logger.info("All motors stopped")

    # ----------------------------------------------------------------------
    def selected_device_changed(self, selected_devices):
        """Called when a device is selected in the BeamlineWidget.
        """

        layout = self._device_widget_layout

        for i in reversed(list(range(layout.count()))):
            item = layout.itemAt(i)
            if item:
                widget = layout.itemAt(i).widget()
                if widget:
                    layout.removeWidget(widget)
                    widget.setVisible(False)
                else:
                    layout.removeItem(item)

        self._selected_devices = selected_devices

        for name in selected_devices:
            for widget in self.device_tuple_map[name]:
                widget.setVisible(True)
                layout.addWidget(widget)
            spacer = QtWidgets.QSpacerItem(22, 20, QtWidgets.QSizePolicy.Expanding,
                                           QtWidgets.QSizePolicy.Expanding)
            layout.addSpacerItem(spacer)

        self.sync_gui("absolute")

    # ----------------------------------------------------------------------
    def _save_snapshot(self):
        """
        """
        # support for non-motor devices? TODO
        self.save_snapshot.emit(json.dumps({"devices": self._selected_devices}))

    # ----------------------------------------------------------------------
    def sync_gui(self, mode="absolute"):
        """
        """
        for _, motors in list(self.device_tuple_map.items()):
            for motor in motors:
                motor.sync_pos(mode, "async")



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'uis/LoggingWidget.ui'
#
# Created by: PyQt5 UI code generator 5.11.3
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_LoggingWidget(object):
    def setupUi(self, LoggingWidget):
        LoggingWidget.setObjectName("LoggingWidget")
        LoggingWidget.resize(562, 627)
        self.verticalLayout = QtWidgets.QVBoxLayout(LoggingWidget)
        self.verticalLayout.setObjectName("verticalLayout")
        self.te_logs = QtWidgets.QTextEdit(LoggingWidget)
        font = QtGui.QFont()
        font.setFamily("Monospace")
        font.setPointSize(10)
        self.te_logs.setFont(font)
        self.te_logs.setReadOnly(True)
        self.te_logs.setTabStopWidth(10)
        self.te_logs.setObjectName("te_logs")
        self.verticalLayout.addWidget(self.te_logs)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem)
        self.label_2 = QtWidgets.QLabel(LoggingWidget)
        self.label_2.setObjectName("label_2")
        self.horizontalLayout_2.addWidget(self.label_2)
        self.cb_log_level = QtWidgets.QComboBox(LoggingWidget)
        self.cb_log_level.setToolTip("")
        self.cb_log_level.setObjectName("cb_log_level")
        self.cb_log_level.addItem("")
        self.cb_log_level.addItem("")
        self.cb_log_level.addItem("")
        self.cb_log_level.addItem("")
        self.horizontalLayout_2.addWidget(self.cb_log_level)
        self.verticalLayout.addLayout(self.horizontalLayout_2)

        self.retranslateUi(LoggingWidget)
        self.cb_log_level.setCurrentIndex(1)
        QtCore.QMetaObject.connectSlotsByName(LoggingWidget)

    def retranslateUi(self, LoggingWidget):
        _translate = QtCore.QCoreApplication.translate
        LoggingWidget.setWindowTitle(_translate("LoggingWidget", "Form"))
        self.label_2.setText(_translate("LoggingWidget", "Log level:"))
        self.cb_log_level.setItemText(0, _translate("LoggingWidget", "Debug"))
        self.cb_log_level.setItemText(1, _translate("LoggingWidget", "Info"))
        self.cb_log_level.setItemText(2, _translate("LoggingWidget", "Warning"))
        self.cb_log_level.setItemText(3, _translate("LoggingWidget", "Error"))




#!/usr/bin/env python

# ----------------------------------------------------------------------
# Author:        Sebastian Piec <sebastian.piec@desy.de>
# Last modified: 2015, April 9
# ----------------------------------------------------------------------

import datetime
import inspect
import logging
import os
import re
import uuid

# only for curr date??? TODO

from PyQt5 import QtWidgets, QtCore

# ----------------------------------------------------------------------
def _makeRange(startStr, endStr, stepStr):
  """
  """
  fromValue = float(startStr)
  toValue = float(endStr)
  valueStep = float(stepStr)

  targetPositions = []

  if abs(valueStep) < 0.0000000001:
    raise RuntimeError("Parameter's step must be non-zero!")
     
  if valueStep > 0:
    while fromValue < toValue:
      targetPositions.append(round(fromValue, 6))     # think on round() TODO!
      fromValue += valueStep
        
  elif valueStep < 0:
    while fromValue > toValue:
      targetPositions.append(round(fromValue, 6))
      fromValue += valueStep

        # include last value? TODO
  if abs(fromValue - toValue) < 0.0000000001:
    targetPositions.append(round(toValue, 6))
   
  return targetPositions



def modificationDate(filename):
  t = os.path.getmtime(filename)
  return datetime.datetime.fromtimestamp(t)



# ----------------------------------------------------------------------
def niceFloat(number):
  """
  """
  strep = str(number)
  return strep.rstrip("0").rstrip(".") if "." in strep else strep



# is postfix used? TODO
# ---------------------------------------------------------------------- 
def generateUniqueName(baseName, usedNames, postfix=""):
  """Generate name unique within "usedNames" name for Region, RegionGroup, etc.
  different from names present in "usedItems".
  """
  baseName += postfix    # e.g. " - Copy"
  newName = baseName
    
  copyCnt = 1
    
  while True:
    isClashing = any([newName == usedName for usedName in usedNames])
      
      # try another name 
    if isClashing:
      newName = baseName + f" ({copyCnt})"
      copyCnt += 1 
    else:
      return newName
       


# ---------------------------------------------------------------------- 
def script_info():
  """ 
  """ 
  from inspect import currentframe, getframeinfo
  
  frameinfo = getframeinfo(currentframe())
  return "%s, line: %d" % (frameinfo.filename, frameinfo.lineno)




# ---------------------------------------------------------------------- 
def makeTimeStamp(dateTime):
  """ Return string representation of QDateTime object.
  """
  format = "yyyy_MM_dd_hhmmss"
  return dateTime.toString(format)
 

# ---------------------------------------------------------------------- 
def _extractRealBase(baseName):
  """ In case there is time stamp, don't append next one...

  tmp_spiec - with this dir will not work...
  """
  tokens = baseName.split("_")
  
    # no timestamp
  if len(tokens) < 5:
    return baseName

  nTokens = len(tokens)

  if re.match(r"\d{6}", tokens[-1]) and re.match(r"\d{2}", tokens[-2]) and \
      re.match(r"\d{2}", tokens[-3]) and re.match(r"\d{4}", tokens[-4]):
      #print (">>>>>>>>>>>>>>>>> FOUND TIME_STAMP, ", tokens[0:nTokens-4])
      return "".join(tokens[0:nTokens-4])

  return baseName
    

      
# ---------------------------------------------------------------------- 
def fileNameWithStamp(fileName, replace=False):
  """If fileName already contains "time stamp" do nothing.
  """
  baseName, fileExtension = os.path.splitext(fileName)

    # possibly substitute previous time stamp
  newBaseName = _extractRealBase(baseName)

    # either "no time stamp" or "user wants to replace it"
  if newBaseName == baseName or replace:
    currentDate = QtCore.QDateTime.currentDateTime()

    fileName = newBaseName + "_" + makeTimeStamp(currentDate) + fileExtension
  
  return fileName
 
 
def makeFileName():
  """
  """ 
 
 
 
# ---------------------------------------------------------------------- 
def niceFileName(fileName, extension):  
  """ Extract file name from a given tuple.
  """
  if not fileName:
    return ""

    # add suffix if it wasn't specified by user
  fileInfo = QtCore.QFileInfo(str(fileName))
  if len(fileInfo.suffix()) == 0:
    fileName = fileInfo.absoluteFilePath() + "." + extension
    
  return fileName


# ---------------------------------------------------------------------- 
def uniqueFileName(baseName, usedNames):
  """Generate unique name for Region, RegionGroup, etc.
  different from names present in "usedItems".
  """
  newName = baseName
  copyCnt = 1
    
  while True:
    clash = False
      
    for usedName in usedNames:
      if newName == usedName:
        clash = True
        break
      
        # newName not found on the list
    if not clash:
      break
     
      # try another name 
    newName = baseName + "_%02d" % copyCnt 
    copyCnt += 1 
    
  return newName




   
# ---------------------------------------------------------------------- 
def parseLogLevel(levelStr):
  """ Return logging level "object" corresponding to a given string 
    representation (case insensitive).
  """
  logLevel = levelStr.lower()
    
  if logLevel == "debug":
    return logging.DEBUG
  elif logLevel == "info":
    return logging.INFO
  elif logLevel == "warning":
    return logging.WARNING
  elif logLevel == "error":
    return logging.ERROR
  elif logLevel == "critical":
    return logging.CRITICAL
  


# ---------------------------------------------------------------------- 
def uniqueName(baseName, postfix, usedItems):
  """Generate unique name for Region, RegionGroup, etc.
  different from names present in "usedItems"."""

  baseName = baseName + postfix    # e.g. " - Copy"
  newName = baseName
    
  copyCnt = 1
    
  while True:
    clash = False
      
    for item in usedItems:
      if newName == item.name:
        clash = True
        break
      
        # newName not found on the list
    if not clash:
      break
     
      # try another name 
    newName = baseName + " (%d)" % copyCnt 
    copyCnt += 1 
    
  return newName





# ----------------------------------------------------------------------  
def assertUniqueInstance(lockFileName, parent):
  """Checks if any other instance of the program is already running on this PC.
  """
  if os.path.isfile(lockFileName):
    msg = f"Looks like another instance of {parent._app_name} is already running!\n"
    msg += "Do you want to continue?"

    response = QtWidgets.QMessageBox.question(parent, "Lockfile found", msg,
                                              QtWidgets.QMessageBox.Yes, QtWidgets.QMessageBox.No)

    if response != QtWidgets.QMessageBox.Yes:
      raise RuntimeError(f"Multiple instance exception ({parent._app_name})")
  
  else:
    with open(lockFileName, "w") as lockFile:
      lockFile.write("locked\n")
      
# ----------------------------------------------------------------------  
def unlockAppInstance(lockFileName):
  """Simply remove "lockfile".
  """
  if os.path.isfile(lockFileName):
    os.remove(lockFileName)
    print(f"Lockfile {lockFileName} removed successfully!")

# ---------------------------------------------------------------------- 
def timeToString(totalAcqTime, options="compact"):
  """Convert time (in seconds) to human readable format.
  """
  secInYear = 365 * 24 * 3600                 # roughly
  secInMonth = 31 * 24 * 3600       
  secInDay = 24 * 3600

  years = int(totalAcqTime / secInYear)
  months = int((totalAcqTime - years * secInYear) / secInMonth)                     # the rest 
  days = int((totalAcqTime - years * secInYear - months * secInMonth) / secInDay)   # --||--

  hours = int((totalAcqTime - years * secInYear - months * secInMonth - days * secInDay) / 3600)
  minutes = int((totalAcqTime - years * secInYear - months * secInMonth - days * secInDay - hours * 3600) / 60)
  seconds = int(totalAcqTime - years * secInYear - months * secInMonth - days * secInDay - hours * 3600 - minutes * 60)
  
  strRep = ""
  if options == "easy":  
    if years != 0:
      strRep = "%dY %dM %dD %dh %dm %ds" % (years, months, days, hours, minutes, seconds)
    elif months != 0:
      strRep = "%dM %dD %dh %dm %ds" % (months, days, hours, minutes, seconds)
    elif days != 0:
      strRep = "%dD %dh %dm %ds" % (days, hours, minutes, seconds)
    elif hours != 0:
      strRep = "%dh %dm %ds" % (hours, minutes, seconds)
    elif minutes != 0:
      strRep = "%dm %ds" % (minutes, seconds)
    elif seconds != 0:
      strRep = "%ds" % (seconds)
  
  else:
    if years != 0:
      strRep = "%dY %dM %dD %02d:%02d:%02d" % (years, months, days, hours, minutes, seconds)
    elif months != 0:
      strRep = "%dM %dD %02d:%02d:%02d" % (months, days, hours, minutes, seconds)
    elif days != 0:
      strRep = "%dD %02d:%02d:%02d" % (days, hours, minutes, seconds)
    else:  
      strRep = "%02d:%02d:%02d" % (hours, minutes, seconds)

  return strRep


# ----------------------------------------------------------------------  
def levelToNumber(levelName):
  """Convert given logging name to integer number.
  """
  levelNumber = 0
  levelName = levelName.lower()

  if levelName == "debug":
    levelNumber = 0 
  elif levelName == "info":
    levelNumber = 1 
  elif levelName == "warning":
    levelNumber = 2 
  elif levelName == "error":
    levelNumber = 3 
    
  return levelNumber

# ----------------------------------------------------------------------
def get_text_coordinates(plot_item, size, xinfo, position='tl'):

    [[x_min, x_max], [y_min, y_max]] = plot_item.viewRange()

    dx = abs(x_max - x_min) * 0.05
    dy = abs(y_max - y_min) * 0.05

    if xinfo == 'binding':
      if 'l' in position:
        position.replace('l', 'r')
      else:
        position.replace('l', 'r')

    if position == 'tl':
      textx = x_min + dx
      texty = y_max - dy
    elif position == 'tr':
      textx = x_max - dx - size.width() * plot_item.getViewBox().viewPixelSize()[0]
      texty = y_max - dy
    elif position == 'bl':
      textx = x_max + dx
      texty = y_min + dy + size.height() * plot_item.getViewBox().viewPixelSize()[1]
    elif position == 'br':
      textx = x_max - dx - size.width() * plot_item.getViewBox().viewPixelSize()[0]
      texty = y_min + dy + size.height() * plot_item.getViewBox().viewPixelSize()[1]

    return QtCore.QPointF(textx, texty)


# ----------------------------------------------------------------------
class Segment:
  """ Simple representation of line segment.
  """

  # ---------------------------------------------------------------------- 
  def __init__(self, x1, y1, x2, y2):
    self.x1 = x1
    self.y1 = y1
    self.x2 = x2
    self.y2 = y2

# ---------------------------------------------------------------------- 
def segmentsCoincident(x1, y1, x2, y2,         # first segment
                       x3, y3, x4, y4):        # second segment
  """
  """
  xMinA = min(x1, x2)
  xMaxA = max(x1, x2)
  yMinA = min(y1, y2)
  yMaxA = max(y1, y2)

  xMinB = min(x3, x4)
  xMaxB = max(x3, x4)
  yMinB = min(y3, y4)
  yMaxB = max(y3, y4)
  
  denominator = (x1 - x2)*(y3 - y4) - (y1 - y2)*(x3 - x4)
  if denominator == 0:      # lines are parallel
      # if any of 2 end points of the first segment is "within" the second segment
    if (xMinA >= xMinB and xMinA <= xMaxB and yMinA >= yMinB and yMinA <= yMaxB) or \
       (xMaxA >= xMinB and xMaxA <= xMaxB and yMaxA >= yMinB and yMaxA <= yMaxB):
      return True

    if (xMinB >= xMinA and xMinB <= xMaxA and yMinB >= yMinA and yMinB <= yMaxA) or \
       (xMaxB >= xMinA and xMaxB <= xMaxA and yMaxB >= yMinA and yMaxB <= yMaxA):
      return True

  return False  

# ---------------------------------------------------------------------- 
def segmentsIntersect(x1, y1, x2, y2,         # first segment
                      x3, y3, x4, y4):        # second segment
  """ Return logical True if given line segments intersect each other.

    The function is used e.g. to verify if motor's trajectory intersects
    with user defined border.
  """
  if segmentsCoincident(x1, y1, x2, y2, x3, y3, x4, y4):
    return True

    # check if segments are parallel but not coincident
  denominator = (x1 - x2)*(y3 - y4) - (y1 - y2)*(x3 - x4)
  if denominator == 0:
    return False

    # calculate intersection point of 2 lines defined by the segments
  px = ((x1*y2 - y1*x2)*(x3 - x4) - (x1 - x2)*(x3*y4 - y3*x4)) / denominator
  py = ((x1*y2 - y1*x2)*(y3 - y4) - (y1 - y2)*(x3*y4 - y3*x4)) / denominator

    # intersection point should be within segmentA and segmentB
  xMinA = min(x1, x2)
  xMaxA = max(x1, x2)
  yMinA = min(y1, y2)
  yMaxA = max(y1, y2)

  xMinB = min(x3, x4)
  xMaxB = max(x3, x4)
  yMinB = min(y3, y4)
  yMaxB = max(y3, y4)
  
  return (px >= xMinA and px <= xMaxA and py >= yMinA and py <= yMaxA and
          px >= xMinB and px <= xMaxB and py >= yMinB and py <= yMaxB)











# yury.matveev@desy.de

"""Simple abstraction on top of the TANGO based stepping motor.

Example usage:
    m = TangoMotor("p09/motor/exp.01")
    m.initialize()

    m.move_absolute(22.1)
    print("Current position:", m.position())

    m.move_relative(-4.5, "async")
"""

import tango
from beamline_console.devices.abstractmotor import AbstractMotor

# ---------------------------------------------------------------------- 
class TangoMotor(AbstractMotor):
    """ 
    """
   
    # ---------------------------------------------------------------------- 
    def __init__(self, server_name="", name="empty"):
        """
        Args:
            (str) TANGO server's name
            (str) user friendly motor's alias
        """
        super().__init__(name)
    
        self.tango_server = str(server_name)
        self.device_proxy = None

    # ----------------------------------------------------------------------
    def state(self):
        """
        Returns:
            (str) current state of the motor as a string
        """
        return str(self.device_proxy.state()).lower()

    # ----------------------------------------------------------------------
    def progress(self):
        """
        Returns:
            (int) current moving progress in range 0-1
        """
        if self.device_proxy.totalmovetime != 0:
            return min(max(1 - self.device_proxy.remainingtime/self.device_proxy.totalmovetime, 0), 1)

    # ----------------------------------------------------------------------
    def position(self):
        """
        Returns:
            (float) current position of the motor
        """
        return self.parameter("position")

    # ----------------------------------------------------------------------
    def move_to_high_limit(self):
        if self.device_proxy.conversion > 0:
            self.device_proxy.movetocwlimit()
        else:
            self.device_proxy.movetoccwlimit()

    # ----------------------------------------------------------------------
    def move_to_low_limit(self):
        if self.device_proxy.conversion > 0:
            self.device_proxy.movetoccwlimit()
        else:
            self.device_proxy.movetocwlimit()

    # ----------------------------------------------------------------------
    def check_limits(self):
        """
        Returns:
            (str) 'high', 'low' or ''
                   if conversion positive, then cwlimit - high, ccwlimit - low and vise versa
         """

        try:
            status = []

            if self.device_proxy.get_property_list("IgnoreLimitSw") and \
                    bool(self.device_proxy.get_property("IgnoreLimitSw")['IgnoreLimitSw'][0]):
                return status

            if self.device_proxy.cwlimit:
                if self.device_proxy.conversion > 0:
                    status.append('high')
                else:
                    status.append('low')

            if self.device_proxy.ccwlimit:
                if self.device_proxy.conversion > 0:
                    status.append('low')
                else:
                    status.append('high')

            return status

        except AttributeError:
            return status

    # ----------------------------------------------------------------------
    def stop(self):
        """Stops the motor, if in the moving state.
        """
        if self.state().lower() == "moving":
            self.device_proxy.stopMove()

    # ----------------------------------------------------------------------
    def calibrate(self, actual_pos):
        """
        Returns:
            actual_pos (float), actual motor's position
        """
        self.device_proxy.command_inout("calibrate", actual_pos)

    # ---------------------------------------------------------------------- 
    def initialize(self):
        """Tango device proxy initialization.
        """
        self.device_proxy = tango.DeviceProxy(self.tango_server)

        elapsed = self.device_proxy.ping()
        print(f"(tangomotor.py) ping {self.tango_server} ({elapsed} us)")

        return True

    # ----------------------------------------------------------------------
    def is_move_valid(self, target_pos):
        """Checks if given target position is within limits.

        Args:
            target_pos (float)
        """
        is_valid, msg = super().is_move_valid(target_pos)
        if not is_valid:
            return False, msg

        else:
            low_limit = self.parameter("UnitLimitMin")        # Tango limits
            high_limit = self.parameter("UnitLimitMax")

            if (target_pos < low_limit or
                target_pos > high_limit):
        
                return False, f"Motor ({self}) target position ({target_pos:.2f}) is "\
                              f"out of range ({low_limit:.2f}, {high_limit:.2f})"

        return True, ""

    # ---------------------------------------------------------------------- 
    def _move_absolute_async(self, target_pos):
        """Asynchronous version of the "move absolute" command. 

        Args:
            target_pos (float)
        """
        is_valid, msg = self.is_move_valid(target_pos)

        if is_valid:
            self.device_proxy.position = target_pos
        else:
            raise ValueError(msg)

    # ----------------------------------------------------------------------
    def status(self):
        """
        Returns:
            (str) status information, e.g. "motor is moving", "idle"
        """
        return str(self.device_proxy.status()).lower()

    # ----------------------------------------------------------------------
    def parameter(self, param_name):
        """
        Returns:
            value of selected motor's parameter (from Tango database).
        """
        return self.device_proxy.read_attribute(param_name).value

    # ----------------------------------------------------------------------
    def set_attribute(self, attribute, value):

        try:
            self.device_proxy.write_attribute(attribute, float(value))
            return True
        except ValueError:
            self.device_proxy.write_attribute(attribute, value)
            return True
        except Exception:
            return False

    # ----------------------------------------------------------------------
    def set_parameter(self, param_name, param_value):
        """Save value of selected motor's property (from Tango database).
        """    
        self.device_proxy.write_attribute(param_name, param_value)

    # ----------------------------------------------------------------------
    def load_xml(self, xml_node):
        """
        """
    # ----------------------------------------------------------------------
    def get_low_limit(self):
        return float(self.parameter("UnitLimitMin"))

    # ----------------------------------------------------------------------
    def get_high_limit(self):
        return float(self.parameter("UnitLimitMax"))


# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'uis/SaveSnapshotDialog.ui'
#
# Created by: PyQt5 UI code generator 5.11.3
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_SaveSnapshotDialog(object):
    def setupUi(self, SaveSnapshotDialog):
        SaveSnapshotDialog.setObjectName("SaveSnapshotDialog")
        SaveSnapshotDialog.resize(569, 574)
        self.verticalLayout = QtWidgets.QVBoxLayout(SaveSnapshotDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.label = QtWidgets.QLabel(SaveSnapshotDialog)
        self.label.setObjectName("label")
        self.horizontalLayout_2.addWidget(self.label)
        self.leSnapshotName = QtWidgets.QLineEdit(SaveSnapshotDialog)
        self.leSnapshotName.setMinimumSize(QtCore.QSize(200, 0))
        self.leSnapshotName.setObjectName("leSnapshotName")
        self.horizontalLayout_2.addWidget(self.leSnapshotName)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem)
        self.btnSelectAll = QtWidgets.QToolButton(SaveSnapshotDialog)
        self.btnSelectAll.setCheckable(False)
        self.btnSelectAll.setObjectName("btnSelectAll")
        self.horizontalLayout_2.addWidget(self.btnSelectAll)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem1)
        self.chbShowAllMotors = QtWidgets.QCheckBox(SaveSnapshotDialog)
        self.chbShowAllMotors.setObjectName("chbShowAllMotors")
        self.horizontalLayout_2.addWidget(self.chbShowAllMotors)
        self.verticalLayout.addLayout(self.horizontalLayout_2)
        self.twDevices = QtWidgets.QTableWidget(SaveSnapshotDialog)
        self.twDevices.setAlternatingRowColors(True)
        self.twDevices.setObjectName("twDevices")
        self.twDevices.setColumnCount(0)
        self.twDevices.setRowCount(0)
        self.verticalLayout.addWidget(self.twDevices)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem2)
        self.btnSaveSnapshot = QtWidgets.QPushButton(SaveSnapshotDialog)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/save.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.btnSaveSnapshot.setIcon(icon)
        self.btnSaveSnapshot.setIconSize(QtCore.QSize(12, 12))
        self.btnSaveSnapshot.setObjectName("btnSaveSnapshot")
        self.horizontalLayout.addWidget(self.btnSaveSnapshot)
        self.btnCancel = QtWidgets.QPushButton(SaveSnapshotDialog)
        self.btnCancel.setObjectName("btnCancel")
        self.horizontalLayout.addWidget(self.btnCancel)
        self.verticalLayout.addLayout(self.horizontalLayout)

        self.retranslateUi(SaveSnapshotDialog)
        self.btnCancel.clicked.connect(SaveSnapshotDialog.reject)
        QtCore.QMetaObject.connectSlotsByName(SaveSnapshotDialog)
        SaveSnapshotDialog.setTabOrder(self.btnSaveSnapshot, self.btnCancel)

    def retranslateUi(self, SaveSnapshotDialog):
        _translate = QtCore.QCoreApplication.translate
        SaveSnapshotDialog.setWindowTitle(_translate("SaveSnapshotDialog", "Save Beamline Snapshot"))
        self.label.setText(_translate("SaveSnapshotDialog", "Name:"))
        self.btnSelectAll.setToolTip(_translate("SaveSnapshotDialog", "Select All Motors"))
        self.btnSelectAll.setText(_translate("SaveSnapshotDialog", "Select All"))
        self.chbShowAllMotors.setText(_translate("SaveSnapshotDialog", "show all motors"))
        self.btnSaveSnapshot.setText(_translate("SaveSnapshotDialog", "Save"))
        self.btnCancel.setText(_translate("SaveSnapshotDialog", "Cancel"))

import beamline_console.gui.icons_rc



"""Displays (most of) the beamline components on a single widget.
"""

from PyQt5 import QtWidgets, QtCore
from beamline_console.base_widget import BaseWidget
from beamline_console.devicecontrol.beamlinegraphicsscene import BeamlineGraphicsScene
from beamline_console.gui.BeamlineWidget_ui import Ui_BeamlineWidget


# ----------------------------------------------------------------------
class BeamlineWidget(BaseWidget):

    WIDGET_NAME = "BeamlineWidget"

    save_snapshot = QtCore.pyqtSignal(str)
    device_selected = QtCore.pyqtSignal(list)

    # ----------------------------------------------------------------------
    def __init__(self, parent):
        """
        Args:
            beamlineHal
            settings
            parent (QWidget)
        """
        BaseWidget.__init__(self, parent)

        self._ui = Ui_BeamlineWidget()
        self._ui.setupUi(self)

        self._access_level = "user"

    # ----------------------------------------------------------------------
    def finish_init(self):

        # widget displaying currently selected devices
        self._ui.device_widget.initialize(self._beamline_hal, self._settings)
        self._ui.device_widget.save_snapshot.connect(lambda selection: self.save_snapshot.emit(selection))
        self._selected_devices = []          # support multi-selection of devices

        self._ui.beamline_view.setFixedHeight(int(self._settings.option("beamline_ui", "section_height"))+25)

        device_list = list(self._beamline_hal.device_map.keys())
        if len(device_list) > 1:
            self._scene = BeamlineGraphicsScene(self._ui.beamline_view, device_list, self._settings)
            self._scene.device_selection_changed.connect(self._ui.device_widget.selected_device_changed)
            self._scene.device_selection_changed.connect(lambda dev_list: self.device_selected.emit(dev_list))
            self._ui.beamline_view.setScene(self._scene)
            self._ui.beamline_view.setDragMode(QtWidgets.QGraphicsView.ScrollHandDrag)
            self._ui.beamline_view.centerOn(0, 0)
            self._ui.but_show_selector.clicked.connect(lambda checked: self._ui.beamline_view.setVisible(checked))
        else:
            self._scene = None
            self._ui.fr_device_selector.setVisible(False)
            self._ui.device_widget.selected_device_changed(device_list)
            self.device_selected.emit(device_list)

        # keep GUI in sync TODO
        self._timer = QtCore.QTimer(self)
        self._timer.timeout.connect(self._synchronize)
        self._timer.start(int(self._settings.option("beamline_ui", "beamline_refresh_period")))

        self._move_tick = 0

    # ----------------------------------------------------------------------
    def access_level_changed(self, access_level):
        """
        """
        self._access_level = access_level

    # ----------------------------------------------------------------------
    def sync_control_widget(self):
        """
        (called e.g. when beamline snapshot is executed)
        """
        self._ui.device_widget.sync_gui()

    # ----------------------------------------------------------------------
    def _synchronize(self):
        """TANGO devices might be moved using other clients like jive...
        """

        self._move_tick += 1

        if self._scene:
            for device in self._scene.device_items:
                if self._move_tick % 2 == 1:
                    self._scene.device_items[device].status_diode.setColor('blue')
                else:
                    there_is_error = False
                    there_is_moving = False
                    try:
                        for motor in self._ui.device_widget.device_tuple_map[device]:
                            motor_state, limits_state = motor.get_status()
                            if motor_state == "fault" or limits_state != []:
                                there_is_error = True
                            elif motor_state == "moving":
                                there_is_moving =  True
                                self._scene.device_items[device].status_diode.setColor('green')
                    except AttributeError:
                        pass
                    finally:
                        if there_is_error:
                            self._scene.device_items[device].status_diode.setColor('ligthred')
                        elif there_is_moving:
                            self._scene.device_items[device].status_diode.setColor('green')



__all__ = ("Release",)

__docformat__ = "restructuredtext"


class Release:
    """Summarize release information as class attributes.
    """

    name = 'beamline_console'
    version_info = (1, 1, 0)
    version = '.'.join(map(str, version_info[:3]))
    release = ''.join(map(str, version_info[3:]))
    separator = '.' if 'dev' in release or 'post' in release else ''
    version_long = version + separator + release

    version_number = int(version.replace('.', ''))
    long_description = description = 'GUI for experimental control'

    license = 'GPL3'
    authors = (('Yury Matveev', 'yury.matveev@desy.de'),)
    author_lines = "\n".join([f"{name}: {email}" for name, email in authors])
    url = 'https://gitlab.desy.de/yury.matveev/beamline_console'
    download_url = 'https://gitlab.desy.de/yury.matveev/beamline_console.git'
    platform = ['Linux', 'Windows']



"""Manage DeviceGraphicsItem objects representing beamline devices.
This class is tightly coupled with BeamlineGraphicsView.
"""

from PyQt5 import QtWidgets, QtCore

from beamline_console.logger.logit import Logit
from beamline_console.devicecontrol.deviceitem.devicegraphicsitem import DeviceGraphicsItem


# ----------------------------------------------------------------------
class BeamlineGraphicsScene(QtWidgets.QGraphicsScene, Logit):
    """
    """
    device_selection_changed = QtCore.pyqtSignal(list)

    # ----------------------------------------------------------------------
    def __init__(self, parent, device_list, settings):


        self._device_list = device_list

        widget_rows = int(settings.option("beamline_ui", "widget_rows"))
        scene_height = int(settings.option("beamline_ui", "section_height"))

        width = len(self._device_list) // widget_rows
        width += 1 if len(self._device_list) % widget_rows else 0
        width *= scene_height

        QtWidgets.QGraphicsScene.__init__(self, QtCore.QRectF(0, 0, width, scene_height), parent)
        Logit.__init__(self)

        self._selected_devices = []
        self.device_items = {}

        horizontal_step = int(settings.option("beamline_ui", "horizontal_step"))
        vertical_step = int(settings.option("beamline_ui", "vertical_step"))
        device_width = int(settings.option("beamline_ui", "device_width"))
        device_height = int(settings.option("beamline_ui", "device_height"))

        shift_x = (horizontal_step - device_width) // 2
        shift_y = (vertical_step - device_height) // 2

        row_count = 0
        column_count = 0
        for device in device_list:
            geometry = [shift_x + column_count*horizontal_step,
                        shift_y + row_count*vertical_step, device_width, device_height]
            self.device_items[device] = DeviceGraphicsItem(self, device, geometry)
            row_count += 1
            if row_count == widget_rows:
                row_count = 0
                column_count += 1
   
    # ----------------------------------------------------------------------
    def mouseMoveEvent(self, event):
        """
        """
        super().mouseMoveEvent(event)
        
        #pos = event.scenePos()

    # ----------------------------------------------------------------------
    def mousePressEvent(self, event):
        """
        """
        super().mousePressEvent(event)

        pos = event.scenePos()

        if QtCore.Qt.ControlModifier != event.modifiers():
            self._selected_devices = []
            
        device_name = self._intersect(pos)
        if device_name:
            if device_name not in self._selected_devices:
                self._selected_devices.append(device_name)
            else:
                self._selected_devices.remove(device_name)

        self.device_selection_changed.emit(self._selected_devices)

    # ----------------------------------------------------------------------
    def _intersect(self, pos):
        """
        """
        for name, graphics_item in list(self.device_items.items()):
            if graphics_item.intersect(pos):
                return name

        return None



# yury.matveev@desy.de

"""Base class for other motor types.
"""

import time
import json


# ---------------------------------------------------------------------- 
class AbstractMotor:
    """
    """

    # ---------------------------------------------------------------------- 
    def __init__(self, name="noname"):
        """
        Args:
            (str) motor's alias
        """
        self.name = name

        # extra software limits (on top of Tango)
        self.user_low_limit = None
        self.user_high_limit = None
        self.decimals = 3
        self._target_pos = 0
        self._access = "user"
        self._is_allowed = True

        self.new_position_when_moving_is_allowed = True

    # ----------------------------------------------------------------------
    def access_level_changed(self, superuser_mode):
        if self._access == "superuser":
            self._is_allowed = superuser_mode

    # ----------------------------------------------------------------------
    def access_guaranteed(self):
        return self._is_allowed

    # ----------------------------------------------------------------------
    def set_access(self, access):
        self._access = access

    # ----------------------------------------------------------------------
    def state(self):
        """
        Returns:
            (str) current state of the motor, e.g.: "on", "moving"
        """
        raise NotImplementedError("Not implemented")

    # ----------------------------------------------------------------------
    def progress(self):
        """
        Returns:
            (int) current moving progress in range 0-1
        """
        raise NotImplementedError("Not implemented")


    # ----------------------------------------------------------------------
    def position(self):
        """
        Returns:
            (float) current position of the motor
        """
        raise NotImplementedError("Not implemented")

    # ---------------------------------------------------------------------- 
    def stop(self):
        """Stops the motor.
        """
        raise NotImplementedError("Not implemented")

    # ---------------------------------------------------------------------- 
    def calibrate(self, actual_pos):
        """Calibrates the motor (actual_pos should be given in units specific
        to the motor).
        """
        raise NotImplementedError("Not implemented")

    # ---------------------------------------------------------------------- 
    def initialize(self):
        """Optional extra initialization.
        """
        return True

    # ----------------------------------------------------------------------
    def move_absolute(self, target_pos, mode="sync"):
        """Moves motor to a given target position. 
        NOTE that the motor will not move if it's in the "moving" state already.
      
        Depending on the "mode" parameter the function will run in either
        synchronous or asynchronous mode.

        Args:
            target_pos (float)
            mode (str): sync, async
        """
        state = self.state()
        if state == "on":
            self._target_pos = target_pos
            if mode.lower() == "sync":
                self._move_absolute_sync(target_pos)
            elif mode.lower() in ["async", 'snapshot']:
                self._move_absolute_async(target_pos)
            else:
                raise RuntimeError(f"Invalid mode '{mode}'")

        elif state == "moving":  # stop and move?
            if self._target_pos != target_pos:
                self.stop()
                self.move_absolute(target_pos, mode)

        else:
            raise RuntimeError("Cannot move motor if it's in '{}'"
                               "state".format(state))

    # ---------------------------------------------------------------------- 
    def move_relative(self, delta, mode="sync"):
        """Moves the motor by a given delta.

        Args:
            delta (float)
            mode (str): sync, asyns
        """
        self.move_absolute(self.position() + delta, mode)

    # ----------------------------------------------------------------------
    def is_move_valid(self, target_pos):
        """Checks if the given target position is within limits.

        Args:
            target_pos (float)
        """
        if not self._is_allowed:
            return False, f"Current user is not allowed to move {self}"

        if self.user_low_limit and target_pos < self.user_low_limit:
            return False, \
                f"motor ({self}), target position ({target_pos:.2f}) below user low limit ({self.user_low_limit:.2f})"

        if self.user_high_limit and target_pos > self.user_high_limit:
            return False, \
                f"motor ({self}), target position ({target_pos:.2f}) below above low limit ({self.user_high_limit:.2f})"

        if not self.new_position_when_moving_is_allowed and self.state() == "moving":
                return False, "motor is moving and new position is not allowed"

        return True, ""

    # ----------------------------------------------------------------------
    def _move_absolute_sync(self, target_pos):
        """Synchronous version of the "move absolute" command.

        Args:
            target_pos (float)
        """
        self._move_absolute_async(target_pos)

        # wait a little bit when the motor is moving
        RATE_SEC = 0.5
        while self.state() == "moving":
            time.sleep(RATE_SEC)

    # ----------------------------------------------------------------------
    def _move_absolute_async(self, target_pos):
        """Asynchronous version of the "move absolute" command
        (control is returned immediately).
    
        Workhorse for other kinds of "move" commands (to be reimplemented in
        derived classes).

        Args:
            target_pos (float)
        """

    # ----------------------------------------------------------------------
    def move_to_high_limit(self):
        raise NotImplementedError("Not implemented")

    # ----------------------------------------------------------------------
    def move_to_low_limit(self):
        raise NotImplementedError("Not implemented")

    # ----------------------------------------------------------------------
    def parameter(self, param_name):
        """
        Returns:
            (float) value of the motor's parameter (delegates request to the
            Tango database typically).
        """
        return 0.0
    
    # ----------------------------------------------------------------------
    def updated_properties(self, node):

        for attribute in list(node._attrs.keys()):
            if (attribute != 'name') and (attribute != 'tango_server'):
                if not hasattr(self, attribute):
                    self.attribute = lambda: None
                try:
                    value = json.loads(node.getAttribute(attribute).lower())
                except:
                    value = node.getAttribute(attribute)
                finally:
                    setattr(self, attribute, value)

    # ----------------------------------------------------------------------
    def set_parameter(self, param_name, param_value):
        """
        """
        raise NotImplementedError("Not implemented")

    # ----------------------------------------------------------------------
    def get_low_limit(self):
        raise NotImplementedError("Not implemented")

    # ----------------------------------------------------------------------
    def get_high_limit(self):
        raise NotImplementedError("Not implemented")

    # ----------------------------------------------------------------------
    def __repr__(self):
        """
        """
        return self.name



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'uis/BeamlineWidget.ui'
#
# Created by: PyQt5 UI code generator 5.11.3
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_BeamlineWidget(object):
    def setupUi(self, BeamlineWidget):
        BeamlineWidget.setObjectName("BeamlineWidget")
        BeamlineWidget.resize(750, 676)
        BeamlineWidget.setMinimumSize(QtCore.QSize(750, 0))
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(BeamlineWidget)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.fr_device_selector = QtWidgets.QFrame(BeamlineWidget)
        self.fr_device_selector.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.fr_device_selector.setFrameShadow(QtWidgets.QFrame.Raised)
        self.fr_device_selector.setObjectName("fr_device_selector")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.fr_device_selector)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.but_show_selector = QtWidgets.QPushButton(self.fr_device_selector)
        font = QtGui.QFont()
        font.setPointSize(15)
        self.but_show_selector.setFont(font)
        self.but_show_selector.setCheckable(True)
        self.but_show_selector.setChecked(True)
        self.but_show_selector.setObjectName("but_show_selector")
        self.verticalLayout.addWidget(self.but_show_selector)
        self.beamline_view = QtWidgets.QGraphicsView(self.fr_device_selector)
        self.beamline_view.setMinimumSize(QtCore.QSize(0, 100))
        self.beamline_view.setMaximumSize(QtCore.QSize(16777215, 1000))
        self.beamline_view.setObjectName("beamline_view")
        self.verticalLayout.addWidget(self.beamline_view)
        self.verticalLayout_2.addWidget(self.fr_device_selector)
        self.device_widget = DeviceWidget(BeamlineWidget)
        self.device_widget.setObjectName("device_widget")
        self.verticalLayout_2.addWidget(self.device_widget)
        self.verticalLayout_2.setStretch(1, 1)

        self.retranslateUi(BeamlineWidget)
        QtCore.QMetaObject.connectSlotsByName(BeamlineWidget)

    def retranslateUi(self, BeamlineWidget):
        _translate = QtCore.QCoreApplication.translate
        BeamlineWidget.setWindowTitle(_translate("BeamlineWidget", "BeamlineWidget"))
        self.but_show_selector.setText(_translate("BeamlineWidget", "Show device selector"))

from beamline_console.devicecontrol.devicewidget import DeviceWidget



# !/usr/bin/python3
import sys, os
from PyQt5 import QtCore, QtGui, QtWidgets


# Originated from
# https://www.mail-archive.com/pyqt@riverbankcomputing.com/msg22889.html
# Modification refered from
# https://gist.github.com/Riateche/27e36977f7d5ea72cf4f

class RangeSlider(QtWidgets.QSlider):
    sliderMoved = QtCore.pyqtSignal(int, int)

    """ A slider for ranges.

        This class provides a dual-slider for ranges, where there is a defined
        maximum and minimum, as is a normal slider, but instead of having a
        single slider value, there are 2 slider values.

        This class emits the same signals as the QSlider base class, with the 
        exception of valueChanged
    """

    # ----------------------------------------------------------------------
    def __init__(self, *args):
        super().__init__(*args)

        self._low = self.minimum()
        self._high = self.maximum()

        self.pressed_control = QtWidgets.QStyle.SC_None
        self.tick_interval = 0
        self.tick_position = QtWidgets.QSlider.NoTicks
        self.hover_control = QtWidgets.QStyle.SC_None
        self.click_offset = 0

        # 0 for the low, 1 for the high, -1 for both
        self.active_slider = 0

    # ----------------------------------------------------------------------
    def low(self):
        return self._low

    # ----------------------------------------------------------------------
    def setLow(self, low: int):
        self._low = low
        self.update()

    # ----------------------------------------------------------------------
    def high(self):
        return self._high

    # ----------------------------------------------------------------------
    def setHigh(self, high):
        self._high = high
        self.update()

    # ----------------------------------------------------------------------
    def paintEvent(self, event):
        # based on http://qt.gitorious.org/qt/qt/blobs/master/src/gui/widgets/qslider.cpp

        painter = QtGui.QPainter(self)
        style = QtWidgets.QApplication.style()

        # draw groove
        opt = QtWidgets.QStyleOptionSlider()
        self.initStyleOption(opt)
        opt.siderValue = 0
        opt.sliderPosition = 0
        opt.subControls = QtWidgets.QStyle.SC_SliderGroove
        if self.tickPosition() != self.NoTicks:
            opt.subControls |= QtWidgets.QStyle.SC_SliderTickmarks
        style.drawComplexControl(QtWidgets.QStyle.CC_Slider, opt, painter, self)
        groove = style.subControlRect(QtWidgets.QStyle.CC_Slider, opt, QtWidgets.QStyle.SC_SliderGroove, self)

        self.initStyleOption(opt)
        opt.subControls = QtWidgets.QStyle.SC_SliderGroove

        opt.siderValue = 0
        opt.sliderPosition = self._low
        low_rect = style.subControlRect(QtWidgets.QStyle.CC_Slider, opt, QtWidgets.QStyle.SC_SliderHandle, self)
        opt.sliderPosition = self._high
        high_rect = style.subControlRect(QtWidgets.QStyle.CC_Slider, opt, QtWidgets.QStyle.SC_SliderHandle, self)

        low_pos = self.__pick(low_rect.center())
        high_pos = self.__pick(high_rect.center())

        min_pos = min(low_pos, high_pos)
        max_pos = max(low_pos, high_pos)

        c = QtCore.QRect(low_rect.center(), high_rect.center()).center()
        if opt.orientation == QtCore.Qt.Horizontal:
            span_rect = QtCore.QRect(QtCore.QPoint(min_pos, c.y() - 2), QtCore.QPoint(max_pos, c.y() + 1))
        else:
            span_rect = QtCore.QRect(QtCore.QPoint(c.x() - 2, min_pos), QtCore.QPoint(c.x() + 1, max_pos))

        if opt.orientation == QtCore.Qt.Horizontal:
            groove.adjust(0, 0, -1, 0)
        else:
            groove.adjust(0, 0, 0, -1)

        highlight = self.palette().color(QtGui.QPalette.Highlight)
        painter.setBrush(QtGui.QBrush(highlight))
        painter.setPen(QtGui.QPen(highlight, 0))

        painter.drawRect(span_rect.intersected(groove))

        for i, value in enumerate([self._low, self._high]):
            opt = QtWidgets.QStyleOptionSlider()
            self.initStyleOption(opt)

            # Only draw the groove for the first slider so it doesn't get drawn
            # on top of the existing ones every time
            if i == 0:
                opt.subControls = QtWidgets.QStyle.SC_SliderHandle  # | QtWidgets.QStyle.SC_SliderGroove
            else:
                opt.subControls = QtWidgets.QStyle.SC_SliderHandle

            if self.tickPosition() != self.NoTicks:
                opt.subControls |= QtWidgets.QStyle.SC_SliderTickmarks

            if self.pressed_control:
                opt.activeSubControls = self.pressed_control
            else:
                opt.activeSubControls = self.hover_control

            opt.sliderPosition = value
            opt.sliderValue = value
            style.drawComplexControl(QtWidgets.QStyle.CC_Slider, opt, painter, self)

    # ----------------------------------------------------------------------
    def __pick(self, pt):
        if self.orientation() == QtCore.Qt.Horizontal:
            return pt.x()
        else:
            return pt.y()

    # # ----------------------------------------------------------------------
    # def mousePressEvent(self, event):
    #     event.accept()
    #
    #     style = QtWidgets.QApplication.style()
    #     button = event.button()
    #
    #     # In a normal slider control, when the user clicks on a point in the
    #     # slider's total range, but not on the slider part of the control the
    #     # control would jump the slider value to where the user clicked.
    #     # For this control, clicks which are not direct hits will slide both
    #     # slider parts
    #
    #     if button:
    #         opt = QtWidgets.QStyleOptionSlider()
    #         self.initStyleOption(opt)
    #
    #         self.active_slider = -1
    #
    #         for i, value in enumerate([self._low, self._high]):
    #             opt.sliderPosition = value
    #             hit = style.hitTestComplexControl(style.CC_Slider, opt, event.pos(), self)
    #             if hit == style.SC_SliderHandle:
    #                 self.active_slider = i
    #                 self.pressed_control = hit
    #
    #                 self.triggerAction(self.SliderMove)
    #                 self.setRepeatAction(self.SliderNoAction)
    #                 self.setSliderDown(True)
    #                 break
    #
    #         if self.active_slider < 0:
    #             self.pressed_control = QtWidgets.QStyle.SC_SliderHandle
    #             self.click_offset = self.__pixelPosToRangeValue(self.__pick(event.pos()))
    #             self.triggerAction(self.SliderMove)
    #             self.setRepeatAction(self.SliderNoAction)
    #     else:
    #         event.ignore()
    #
    # # ----------------------------------------------------------------------
    # def mouseMoveEvent(self, event):
    #     if self.pressed_control != QtWidgets.QStyle.SC_SliderHandle:
    #         event.ignore()
    #         return
    #
    #     event.accept()
    #     new_pos = self.__pixelPosToRangeValue(self.__pick(event.pos()))
    #     opt = QtWidgets.QStyleOptionSlider()
    #     self.initStyleOption(opt)
    #
    #     if self.active_slider < 0:
    #         offset = new_pos - self.click_offset
    #         self._high += offset
    #         self._low += offset
    #         if self._low < self.minimum():
    #             diff = self.minimum() - self._low
    #             self._low += diff
    #             self._high += diff
    #         if self._high > self.maximum():
    #             diff = self.maximum() - self._high
    #             self._low += diff
    #             self._high += diff
    #     elif self.active_slider == 0:
    #         if new_pos >= self._high:
    #             new_pos = self._high - 1
    #         self._low = new_pos
    #     else:
    #         if new_pos <= self._low:
    #             new_pos = self._low + 1
    #         self._high = new_pos
    #
    #     self.click_offset = new_pos
    #
    #     self.update()
    #
    #     # self.emit(QtCore.SIGNAL('sliderMoved(int)'), new_pos)
    #     self.sliderMoved.emit(self._low, self._high)
    #
    # # ----------------------------------------------------------------------
    # def __pixelPosToRangeValue(self, pos):
    #     opt = QtWidgets.QStyleOptionSlider()
    #     self.initStyleOption(opt)
    #     style = QtWidgets.QApplication.style()
    #
    #     gr = style.subControlRect(style.CC_Slider, opt, style.SC_SliderGroove, self)
    #     sr = style.subControlRect(style.CC_Slider, opt, style.SC_SliderHandle, self)
    #
    #     if self.orientation() == QtCore.Qt.Horizontal:
    #         slider_length = sr.width()
    #         slider_min = gr.x()
    #         slider_max = gr.right() - slider_length + 1
    #     else:
    #         slider_length = sr.height()
    #         slider_min = gr.y()
    #         slider_max = gr.bottom() - slider_length + 1
    #
    #     return style.sliderValueFromPosition(self.minimum(), self.maximum(),
    #                                          pos - slider_min, slider_max - slider_min,
    #                                          opt.upsideDown)


#!/usr/bin/env python

# ----------------------------------------------------------------------
# Author:   spiec@
# Modified: 2019
# ----------------------------------------------------------------------
import os
import logging

from PyQt5 import QtCore, QtGui

APP_NAME = 'ExperimentalControl'
SCRIPT_LOG = 'ExperimentalControlScript'

LOG_FOLDER = f"{os.path.expanduser('~')}/.beamline_console/logs"
LOG_FORMATTER = logging.Formatter("%(asctime)s %(filename)s:%(lineno)d %(levelname)-8s %(message)s")

# Log widget coloring

DEBUG_COLOR = QtGui.QColor(0, 0, 0)
INFO_COLOR = QtGui.QColor(80, 230, 80)
WARNING_COLOR = QtGui.QColor(255, 93, 42)
ERROR_COLOR = QtGui.QColor(255, 42, 42)

DATE_COLOR = QtGui.QColor(90, 90, 90)

DEVICE_TITEL = QtGui.QFont("Bitstream Charter", 18, QtGui.QFont.Bold)


"""Show historical positions of the device.
"""

from PyQt5 import QtWidgets

from beamline_console.logger.logit import Logit
from beamline_console.ui_experimentcontrol.DeviceHistoryDialog_ui import Ui_DeviceHistoryDialog


# ----------------------------------------------------------------------
class DeviceHistoryDialog(QtWidgets.QDialog, Logit):
    """
    """
    TIMESTAMP_STYLE = "color: #777777;"

    # ----------------------------------------------------------------------
    def __init__(self, deviceName, beamlineHal, parent):
        """
        Args:
        """
        QtWidgets.QDialog.__init__(self, parent)
        Logit.__init__(self)

        self._ui = Ui_DeviceHistoryDialog()
        self._ui.setupUi(self)

        msg = ""
        for entry in reversed(beamlineHal.history):
            if entry[0] == deviceName:
                msg += self._makeEntry(entry)

        self._ui.teHistory.setText(msg)

        self.setWindowTitle(f"{deviceName} logs")

    # ----------------------------------------------------------------------
    def _makeEntry(self, entry):
        """
        """
        return "<span style=\"{}\">[{}]</span> {}/{} moved from {:.2f} to {:.2f}<br>".format(
            self.TIMESTAMP_STYLE, entry[-1], *entry[:-1])




import sys
import os
import logging
import traceback
import time
import shutil

try:
    from io import StringIO
except ImportError:
    from io import StringIO

from pathlib import Path
from PyQt5 import QtWidgets, QtCore
from logging.handlers import RotatingFileHandler

from optparse import OptionParser
from beamline_console.constants import APP_NAME, LOG_FORMATTER, LOG_FOLDER
from beamline_console.main_window import ExperimentalControl


# ----------------------------------------------------------------------
def abspath(*path):
    """A method to determine absolute path for a given relative path to the
    directory where this setup.py script is located"""
    setup_dir = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(setup_dir, *path)


# ----------------------------------------------------------------------
def check_installation():
    home_path = os.path.join(os.path.expanduser('~'), '.beamline_console')
    if not os.path.exists(home_path):
        os.mkdir(home_path)

        # copy default settings
        shutil.copyfile(abspath('default_config/main.cfg'), os.path.join(home_path, 'main.cfg'))

        # making xml folder
        shutil.copyfile(abspath('default_config/devices.xml'), os.path.join(home_path, 'devices.xml'))

        snapshot_folder = os.path.join(home_path, 'bl_snapshots')
        os.mkdir(snapshot_folder)

# --------------------------------------------------------------------
def excepthook(exc_type, exc_value, traceback_obj):
    """
    Global function to catch unhandled exceptions. This function will result in an error dialog which displays the
    error information.

    :param exc_type: exception type
    :param exc_value: exception value
    :param traceback_obj: traceback object
    :return:
    """
    separator = '-' * 80
    log_path = f"{Path.home()}/.beamline_console/logs/error.log"
    time_string = time.strftime("%Y-%m-%d, %H:%M:%S")
    tb_info_file = StringIO()
    traceback.print_tb(traceback_obj, None, tb_info_file)
    tb_info_file.seek(0)
    tb_info = tb_info_file.read()
    errmsg = f'{str(exc_type)}: \n{str(exc_value)}'
    sections = [separator, time_string, separator, errmsg, separator, tb_info]
    msg = '\n'.join(sections)
    try:
        f = open(log_path, "a")
        f.write(msg)
        f.close()
    except OSError:
        pass

    msg_box = QtWidgets.QMessageBox()
    msg_box.setModal(False)
    msg_box.setIcon(QtWidgets.QMessageBox.Critical)
    msg_box.setText(msg)
    msg_box.setInformativeText(msg)
    msg_box.setWindowTitle("Error")
    msg_box.setStandardButtons(QtWidgets.QMessageBox.Ok)
    msg_box.show()


# --------------------------------------------------------------------
def setup_logger(args):

    while not os.path.exists(LOG_FOLDER):
        folder = LOG_FOLDER
        while not os.path.exists(folder):
            try:
                os.mkdir(folder)
            except FileNotFoundError:
                folder = os.path.dirname(folder)

    log_level = logging.DEBUG

    filename = os.path.join(LOG_FOLDER, "main.log")
    print(f"Main logs to file: {filename}")

    main_handler = RotatingFileHandler(filename, mode='a', maxBytes=5 * 1024 * 1024,
                                     backupCount=2, encoding=None, delay=0)
    main_handler.setFormatter(LOG_FORMATTER)
    main_handler.setLevel(log_level)

    app_log = logging.getLogger(APP_NAME)
    app_log.setLevel(log_level)

    app_log.addHandler(main_handler)

    if args.log:
        console = logging.StreamHandler()
        console.setLevel(log_level)
        console.setFormatter(LOG_FORMATTER)
        app_log.addHandler(console)


# --------------------------------------------------------------------
def get_options(args=None):
    parser = OptionParser()

    parser.add_option("-u", "--user", dest='user', default='user')
    parser.add_option("--log", action='store_true', dest='log', help="print logs to console")

    (options, _) = parser.parse_args(args)

    return options


# --------------------------------------------------------------------
def main():

    check_installation()

    args = get_options(sys.argv)
    setup_logger(args)

    app = QtWidgets.QApplication([])
    sys.excepthook = excepthook

    mainWindow = ExperimentalControl(args)
    if mainWindow.finish_init():
        mainWindow.show()
        app.exec_()
    else:
        mainWindow.exit_program(True)

# --------------------------------------------------------------------
# Start Qt event loop unless running in interactive mode.
if __name__ == '__main__':
    main()



from datetime import datetime
import re
import logging
import sys
import threading
import time

from beamline_console.constants import APP_NAME

from xml.dom.minidom import (Document, parseString)

logger = logging.getLogger(APP_NAME)


# ----------------------------------------------------------------------
class BeamlineSnapshot(threading.Thread):
    """
    """
    TIME_STAMP_FMT = "%Y-%m-%d %H:%M:%S"
    REFRESH_RATE = 1.5            # seconds

    # ----------------------------------------------------------------------
    def __init__(self, beamline_hal=None, filename="", name="",):
        """
        """
        super().__init__()

        self.filename = filename
        self.name = name

        self.beamline_hal = beamline_hal

        self.motor_items = []           # list of tuples
   
        self.date_time = ""

        if self.filename:
            self.load(self.filename)    # can throw

        self.break_loop = False
        self.state = "idle"

        self.exc_type = None
        self.exc_value = None
    
    # ----------------------------------------------------------------------
    def add_item(self, device_name, motor_name, position, step_number):
        """
        """
        self.motor_items.append([device_name, motor_name, position, step_number])

    # ----------------------------------------------------------------------
    def update_position(self):
        for item in self.motor_items:
            full_name = f"{item[0]}/{item[1]}"
            item[2] = self.beamline_hal.motor_position(full_name)

    # ----------------------------------------------------------------------
    def validate(self):
        """Check if motors from this snapshot can be moved to requested target
        positions.
        """
        for item in self.motor_items:
            full_name = f"{item[0]}/{item[1]}"
            target_pos = float(item[2])

            flag, msg = self.beamline_hal.is_move_valid(full_name, target_pos)
            if not flag:
                logger.error(msg)
                raise RuntimeError(msg)
  
    # ----------------------------------------------------------------------
    def run(self):
        """Move motors to positions save in the snapshot.
        """
        try:
            self._execute()

        except Exception as exc:
            self.exc_type, self.exc_value = sys.exc_info()[:2]   # for thread's caller
            logger.exception(exc)
            raise

    # ----------------------------------------------------------------------
    def _execute(self):
        """
        """
        self.break_loop = False
        self.state = "running"

        logger.info(f"Moving to snapshot {self.name}")

        move_groups = {}
        for item in self.motor_items:
            step = item[3]

            if not step in move_groups:
                move_groups[step] = []

            move_groups[step].append(item[:3])

            # move motors in groups
        for group_idx, move_group in list(move_groups.items()):
            logger.debug(f"START mv group {group_idx}")
            
            n_motors = len(move_group)
            for motor_idx in range(n_motors):
                dev_name, motor_name, position = move_group[motor_idx]
      
                    # TODO 
                    # all motors or none should be moved! 
                try:
                    full_name = f"{dev_name}/{motor_name}"
                    self.beamline_hal.move_motor(full_name, position,
                                                 "absolute", "snapshot")
        
                except RuntimeError as err:   #, PyTango.DevFailed) as err:
                    self.beamline_hal.group_action("stop_motors")
                    raise       # requires more work if one wants to transfer exception to higher level TODO
      
            while True:
                time.sleep(self.REFRESH_RATE)

                if self.break_loop:
                    self.set_state("finished")
                    return

                if self.state in ["idle", "paused"]:
                    logger.debug(f"snapshot state {self.state}")
                    continue

                    # wait until all motors are moved to target positions
                n_finished = 0
                for motor_idx in range(n_motors):  
                    full_name = f"{dev_name}/{motor_name}"
                    dev_name, motor_name = move_group[motor_idx][:2]

                    if self.beamline_hal.motor_state(full_name) != "moving":
                        n_finished += 1
       
                if n_finished == n_motors:
                    logger.debug(f"group {group_idx} finished")
                    break
    
        self.set_state("finished")

        logger.info(f"Snapshot {self.name} completed")

    # ----------------------------------------------------------------------
    def abort(self):
        """
        """
        self.break_loop = True
        self.set_state("finished")

    # ----------------------------------------------------------------------
    def pause(self):
        """
        """
            #self.pause_movement() TODO
        self.set_state("paused")

    # ----------------------------------------------------------------------
    def resume(self):
        """
        """
        self.set_state("running")

    # ----------------------------------------------------------------------
    def load(self, filename):
        """Load snapshot from a given file.
    
        Args:
            filename (str)
        """
        self._reset()

        self.filename = str(filename)
        with open(self.filename) as f:
            content = f.read()
   
        dom = parseString(content)        # DOM
      
        nodes = dom.getElementsByTagName("beamline_snapshot") 
        try:
            root_node = nodes[0]

            self.name = root_node.getAttribute("name")
            self.date_time = root_node.getAttribute("datetime")

            item_nodes = root_node.getElementsByTagName("item")

            for item in item_nodes:
                device_name = item.getAttribute("device_name")
                motor_name  = item.getAttribute("motor_name")
                position    = float(item.getAttribute("target_position"))
                step_number = int(item.getAttribute("step_number"))
        
                self.add_item(device_name, motor_name, position, step_number) 
    
        except Exception as err:
            self.exc_type, self.exc_value = sys.exc_info()[:2]
     
        return len(self.motor_items) > 0

    # ----------------------------------------------------------------------
    def save(self, filename=""):
        """
        """
        if filename:
            self.filename = str(filename)

        doc = Document()

        root_node = doc.createElement("beamline_snapshot")
        doc.appendChild(root_node)
        root_node.setAttribute("name", self.name)
      
        self.date_time = datetime.now().strftime(self.TIME_STAMP_FMT)
        root_node.setAttribute("datetime", self.date_time)

        for item in self.motor_items:
            node = doc.createElement("item")
            root_node.appendChild(node)
      
            node.setAttribute("device_name", str(item[0]))
            node.setAttribute("motor_name", str(item[1]))
            node.setAttribute("target_position", str(item[2]))
            node.setAttribute("step_number", str(item[3]))

        with open(self.filename, "w") as f:
            f.write(doc.toprettyxml(indent="  "))

    # ----------------------------------------------------------------------
    def last_exception(self):
        """
        """
        exc_type, exc_value = self.exc_type, self.exc_value
        self.exc_type, self.exc_value = None, None

        return exc_type, exc_value

    # ----------------------------------------------------------------------
    def empty(self):
        """
        """
        return len(self.motor_items) < 1

    # ----------------------------------------------------------------------
    def set_state(self, new_state):
        """
        """
        self.state = new_state

    # ----------------------------------------------------------------------
    def _reset(self):
        """
        """
        self.filename = ""
        self.name = ""
        self.motor_items = []
   
        self.exc_value = None
        self.exc_type = None
    
        self.date_time = ""

        self.break_loop = False
        self.state = "idle"

    # ----------------------------------------------------------------------
    @staticmethod
    def is_valid_name(snapshot_name):
        """Validate given snapshot name.

        Returns:
            (bool)
        """
        if not snapshot_name:
            return False

        if 'global' in snapshot_name:
            return False
        
        return re.match(r"^[a-zA-Z0-9_\-=]+$", snapshot_name) is not None




"""Device status diode
"""

from PyQt5 import QtWidgets, QtCore, QtGui

from beamline_console.logger.logit import Logit


# ----------------------------------------------------------------------
class StatusItem(QtWidgets.QGraphicsRectItem, Logit):

        # adjust to your needs and taste
    SELECTED_STYLE = {"pen": QtGui.QPen(QtGui.QBrush(QtGui.QColor(50, 205, 50)), 2,
                                        QtCore.Qt.SolidLine, QtCore.Qt.SquareCap,
                                        QtCore.Qt.BevelJoin),
                      "brush": QtGui.QBrush(QtGui.QColor(10, 30, 190, 10))
                     }
                      
    DESELECTED_STYLE = {"pen": QtGui.QPen(QtGui.QBrush(QtGui.QColor(200, 200, 200)), 1,
                                          QtCore.Qt.SolidLine, QtCore.Qt.SquareCap,
                                          QtCore.Qt.BevelJoin),
                        "brush": QtGui.QBrush(QtGui.QColor(10, 30, 190, 10))
                       }

    WIDTH = 25
    HEIGHT = 18

    # ----------------------------------------------------------------------
    def __init__(self, frame_geometry):
        QtWidgets.QGraphicsRectItem.__init__(self)
        Logit.__init__(self)
       
        self._initial_geom = QtCore.QRectF(*frame_geometry)
        self._frame_geometry = frame_geometry
        x = frame_geometry[0] + frame_geometry[3] - self.WIDTH   #+ frameGeometry[2] - self.WIDTH
        y = frame_geometry[1] - 25 + 18 #headerPos
        
        self.setRect(x, y + 8, self.WIDTH, frame_geometry[3] - 1)

        rect = self.boundingRect()
        lgrad = QtGui.QLinearGradient(rect.topLeft(), rect.topRight())
        lgrad.setColorAt(1.0, QtGui.QColor(30, 144, 255, 200))
        lgrad.setColorAt(0.0, QtGui.QColor(30, 144, 255, 0))
        self.setBrush(lgrad)

        self.setPen(QtGui.QPen(QtGui.QColor(0, 0, 0, 0)))

    # ----------------------------------------------------------------------
    def setActive(self, isActive):
        """
        """
        #self.selRect.setPen(self.ACTIVE_PEN if isActive else self.INACTIVE_PEN)
        
    # ----------------------------------------------------------------------
    def updateOffsetX(self, offsetX):
        """
        """
        self.offsetX = offsetX
        
        rect = self.rect()
        #rect.setLeft(self._initialGeom.left() + self._initialGeom.width() - self.WIDTH + self.offsetX)
        rect.setLeft(self._initial_geom.left() + self.offsetX + self._frame_geometry[2] - self.WIDTH)
        rect.setRight(self._initial_geom.right() - self._initial_geom.width() + self.offsetX + self._frame_geometry[2])
        self.setRect(rect)

    # ----------------------------------------------------------------------
    def setColor(self, color):
        rect=self.rect()
        lgrad = QtGui.QLinearGradient(rect.topLeft(), rect.topRight())
        lgrad.setColorAt(0.0, QtGui.QColor(30, 144, 255, 0))

        if color =='strongred':
            self.setBrush(QtGui.QBrush(QtGui.QColor(30, 144, 255, 70))) #QtGui.QColor(255, 10, 10, 190)
            self.setPen(QtGui.QPen(QtGui.QColor(30, 144, 255, 120))) #QtGui.QColor(255, 10, 10, 255)
        elif color == 'ligthred':
            lgrad.setColorAt(1.0, QtGui.QColor(205, 10, 10, 150))
            self.setBrush(lgrad)

        elif color =='green':
            lgrad.setColorAt(1.0, QtGui.QColor(50, 205, 50, 190))
            self.setBrush(lgrad)

        elif color =='blue':
            lgrad.setColorAt(1.0, QtGui.QColor(30, 144, 255, 120))
            self.setBrush(lgrad)

            #self.setBrush(QtGui.QBrush(QtGui.QColor(30, 144, 255, 70))) #50, 205, 50, 190

            #
        elif color == 'orange':
            self.setBrush(QtGui.QBrush(QtGui.QColor(205, 205, 50, 190)))





# yury.matveev@desy.de

"""
Widget displaying saved beamline snapshots.
"""

import os
import sys

from distutils.util import strtobool

from PyQt5 import QtWidgets, QtCore, QtGui

from beamline_console.base_widget import BaseWidget
from beamline_console.snapshot.snapshotmanager import SnapshotManager
from beamline_console.snapshot.beamsnapshot import BeamlineSnapshot
from beamline_console.snapshot.savesnapshotdialog import SaveSnapshotDialog
from beamline_console.gui.SnapshotWidget_ui import Ui_SnapshotWidget


# ----------------------------------------------------------------------
class SnapshotWidget(BaseWidget):
    """
    """
    WIDGET_NAME = "SnapshotWidget"

    motors_moved = QtCore.pyqtSignal()

    # ----------------------------------------------------------------------
    def __init__(self, parent):
        BaseWidget.__init__(self, parent)

        self._ui = Ui_SnapshotWidget()
        self._ui.setupUi(self)

        self._snapshot_manager = SnapshotManager(parent.beamline_hal)

        self._snapshot_dir = ""
        self.device_selected([""])

        self._access_level = "user"

        self._ui.tw_snapshots.cellDoubleClicked.connect(self._use_selected_snapshot)
        self._ui.but_apply.clicked.connect(self._use_selected_snapshot)
        self._ui.but_delete.clicked.connect(self._delete_selected_snapshot)

        self._context_menu = QtWidgets.QMenu("SnapshotsWidget_ctxmenu", self)

        act_use_snapshot = QtWidgets.QAction("Use Snapshot", self)
        act_use_snapshot.triggered.connect(self._use_selected_snapshot)
        self._context_menu.addAction(act_use_snapshot)

        self.act_update_snapshot = QtWidgets.QAction("Update Snapshot", self)
        self.act_update_snapshot.triggered.connect(self._update_selected_snapshot)
        self.act_update_snapshot.setVisible(False)
        self._context_menu.addAction(self.act_update_snapshot)

        act_delete_snapshot = QtWidgets.QAction("Delete Snapshot", self)
        act_delete_snapshot.triggered.connect(self._delete_selected_snapshot)
        self._context_menu.addAction(act_delete_snapshot)

        act_refresh = QtWidgets.QAction(QtGui.QIcon(":/icons/refresh.png"), "Refresh", self)
        act_refresh.triggered.connect(self.refresh)

        self._context_menu.addSeparator()
        self._context_menu.addAction(act_refresh)

        self._ui.tw_snapshots.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self._ui.tw_snapshots.customContextMenuRequested.connect(lambda:
            self._context_menu.exec_(QtGui.QCursor.pos()))

        self.setContextMenuPolicy(QtCore.Qt.NoContextMenu)

    # ----------------------------------------------------------------------
    def refresh(self):
        """
        """
        self._snapshot_manager.load(self._snapshot_dir)
        self._ui.tw_snapshots.setRowCount(0)

        headers = ["Name"]

        if self._snapshot_manager.beamline_snapshots:
            for ind, motor in enumerate(self._snapshot_manager.beamline_snapshots[0].motor_items):
               headers.append(motor[1])

        # headers += [""]

        self._ui.tw_snapshots.setColumnCount(len(headers))
        self._ui.tw_snapshots.setHorizontalHeaderLabels(headers)

        for idx, snapshot in enumerate(self._snapshot_manager.beamline_snapshots):
            self._ui.tw_snapshots.setRowCount(idx + 1)
            self._ui.tw_snapshots.setItem(idx, 0, QtWidgets.QTableWidgetItem(snapshot.name))
            for ind, motor in enumerate(snapshot.motor_items):
                self._ui.tw_snapshots.setItem(idx, 1 + ind, QtWidgets.QTableWidgetItem(str(motor[2])))

        header = self._ui.tw_snapshots.horizontalHeader()
        header.setSectionResizeMode(0, QtWidgets.QHeaderView.Stretch)
        for ind in range(1, len(headers)):
            header.setSectionResizeMode(ind , QtWidgets.QHeaderView.ResizeToContents)

    # ----------------------------------------------------------------------
    def access_level_changed(self, access_level):
        """
        """
        self._access_level = access_level
        self.act_update_snapshot.setVisible(access_level=="superuser")

    # ----------------------------------------------------------------------
    def device_selected(self, devices):
        app_dir = os.path.join(os.path.join(os.path.expanduser('~'), '.beamline_console'))
        if "bl_snapshots" not in os.listdir(app_dir):
            os.mkdir(os.path.join(app_dir, "bl_snapshots"))

        self._snapshot_dir = os.path.join(app_dir, "bl_snapshots")

        device = devices[0]
        if device:
            if device not in os.listdir(self._snapshot_dir):
                os.mkdir(os.path.join(self._snapshot_dir, device))

            self._snapshot_dir = os.path.join(self._snapshot_dir, device)

        self.refresh()

    # ----------------------------------------------------------------------
    def _update_selected_snapshot(self):
        if self._access_level == "superuser":
            items = self._ui.tw_snapshots.selectedItems()
            if items:
                snapshot_name = items[0].text()
                msg = f"Are you sure you want to update snapshot {snapshot_name}?"
                if QtWidgets.QMessageBox.question(self, "Confirm", msg,
                                              QtWidgets.QMessageBox.Yes,
                                              QtWidgets.QMessageBox.No) == QtWidgets.QMessageBox.Yes:

                    self._snapshot_manager.update_snapshot(snapshot_name)
                    self.refresh()

    # ----------------------------------------------------------------------
    def _use_selected_snapshot(self):
        items = self._ui.tw_snapshots.selectedItems()
        if items:
            filename = f"{os.path.join(self._snapshot_dir, items[0].text())}.xml"

            try:
                if strtobool(self._settings.option("snapshots", "confirmation_to_apply")):
                    response, move_scheme = self._confirm_move(filename)
                else:
                    response = True

            except Exception as err:
                self.logger.error(f"Cannot confirm move: {repr(err)}", exc_info=sys.exc_info())
                raise
            else:
                if response:
                    try:
                        self._beamline_hal.move_snapshot(filename, "async")
                    except Exception as err:
                        self.logger.error(f"Error during move: {repr(err)}", exc_info=sys.exc_info())
                        raise

    # ----------------------------------------------------------------------
    def _delete_selected_snapshot(self):
        items = self._ui.tw_snapshots.selectedItems()
        if items:
            snapshot_name= items[0].text()

            msg = f"Are you sure you want to delete snapshot {snapshot_name}?"
            if QtWidgets.QMessageBox.question(self, "Confirm", msg,
                                          QtWidgets.QMessageBox.Yes,
                                          QtWidgets.QMessageBox.No) == QtWidgets.QMessageBox.Yes:

                self._snapshot_manager.delete_snapshot(snapshot_name)
                self.refresh()

    # ----------------------------------------------------------------------
    def _confirm_move(self, filename):
        """
        Args:
            filename (str)
        Returns:
            flag
            move_map
        """
        move_groups = {}

        snapshot = BeamlineSnapshot(filename)
        for item in snapshot.motor_items:
            step = item[3]

            if not step in move_groups:
                move_groups[step] = []

            move_groups[step].append(item[:3])

        move_map = {"status": "ok",
                   "type": "move_motor",
                   "details": []}

        scheme = ""
        for group_idx, (group_name, group_handle) in enumerate(move_groups.items()):
            scheme += f"Step {group_idx}\n"

            for item in group_handle:
                scheme += f"   {item[0]}/{item[1]}, target: {float(item[2]):.5f}\n"

                motor_name = "{}/{}".format(*item[:2])

                prev_pos = self._beamline_hal.motor_position(motor_name)

                move_map["details"].append({"motor_name": motor_name,
                                           "prev_pos": prev_pos,
                                           "target_pos": item[2]
                                          })

        msg = "Use snapshot {} created on {}?\n\n{}".format(filename,
                                                            snapshot.date_time,
                                                            scheme)
        msg_box = QtWidgets.QMessageBox()
        msg_box.setText(msg)
        msg_box.setStandardButtons(QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
        msg_box.setWindowTitle("Confirm")

        return (msg_box.exec_() == QtWidgets.QMessageBox.Yes), move_map

    # ----------------------------------------------------------------------
    def save_snapshot(self, selection):
        """
        Args:
            (map) selection, e.g. {"device": ["table", "manipulator"]}
        """
        SaveSnapshotDialog(self._beamline_hal, selection, self._snapshot_manager).exec_()
        self.refresh()



import json
import os
import logging

from PyQt5 import QtWidgets, QtGui

from beamline_console.constants import APP_NAME
from beamline_console.snapshot.beamsnapshot import BeamlineSnapshot
from beamline_console.devices.motordevice import MotorBasedDevice
from beamline_console.gui.SaveSnapshotDialog_ui import Ui_SaveSnapshotDialog

logger = logging.getLogger(APP_NAME)


# ----------------------------------------------------------------------
class SaveSnapshotDialog(QtWidgets.QDialog):

    # ----------------------------------------------------------------------
    def __init__(self, beamline_hal, selection, snapshot_manager, parent=None):
        super().__init__(parent)

        self.beamline_hal = beamline_hal
        self.snapshot_manager = snapshot_manager
        
        self.device_selection = None
        if selection is not None:
            self.device_selection = [str(name) for name in json.loads(str(selection))["devices"]]

        self.columns = ["Motor Name", "Save", "Step"]
        self.motor_tuples = []

        self._ui = Ui_SaveSnapshotDialog()
        self._ui.setupUi(self)

        self._ui.btnSaveSnapshot.clicked.connect(self.save_snapshot)
        self._ui.btnSelectAll.clicked.connect(self.select_all)
        self._ui.chbShowAllMotors.stateChanged.connect(self.show_all_toggled)

        self.show_motors(self.device_selection)

        self.select_all()

    # ----------------------------------------------------------------------
    def show_motors(self, selection=None):
        # called when devicewidget sends save_snapshot signal
        self._ui.twDevices.clear()
        self._ui.twDevices.setColumnCount(len(self.columns))
        self._ui.twDevices.setHorizontalHeaderLabels(self.columns)

        self.motor_tuples = []
        for dev_idx, (dev_name, dev_handle) in enumerate(self.beamline_hal.device_map.items()):
            if (isinstance(dev_handle, MotorBasedDevice) and
                (selection is None or dev_name in selection)):
                for motor in list(dev_handle.motors.values()):
                    chb_save = QtWidgets.QCheckBox(self._ui.twDevices)
                    sb_step = QtWidgets.QSpinBox(self._ui.twDevices)
                    sb_step.setEnabled(False)
                    sb_step.setRange(1, 1000)
                    sb_step.setValue(1)
                    chb_save.toggled.connect(sb_step.setEnabled)

                    self.motor_tuples.append((dev_name, motor.name,
                                              chb_save, sb_step, self._next_color(dev_idx)))

        self._ui.twDevices.setRowCount(len(self.motor_tuples))

        for motor_idx, motor_tuple in enumerate(self.motor_tuples):
            #name_item = QtWidgets.QTableWidgetItem(motor_tuple[1])
            name_item = QtWidgets.QTableWidgetItem("{}/{}".format(*motor_tuple[:2]))    # full name?
            name_item.setBackground(QtGui.QBrush(QtGui.QColor(*motor_tuple[4])))
            
            self._ui.twDevices.setItem(motor_idx, 0, name_item)
            self._ui.twDevices.setCellWidget(motor_idx, 1, motor_tuple[2])
            self._ui.twDevices.setCellWidget(motor_idx, 2, motor_tuple[3])

        self._ui.twDevices.resizeColumnsToContents()
        self._ui.twDevices.resizeRowsToContents()
        self._ui.twDevices.setColumnWidth(0, 180)

    # ----------------------------------------------------------------------
    def show_all_toggled(self, flag):
        self.show_motors(None if flag else self.device_selection)

    # ----------------------------------------------------------------------
    def select_all(self):
        n_selected = 0
        for row in range(self._ui.twDevices.rowCount()):
            n_selected += 1 if self._ui.twDevices.cellWidget(row, 1).isChecked() else 0

        flag = not (n_selected == self._ui.twDevices.rowCount())
        for row in range(self._ui.twDevices.rowCount()):
            self._ui.twDevices.cellWidget(row, 1).setChecked(flag)

    # ----------------------------------------------------------------------
    def save_snapshot(self):
        snapshot_name = str(self._ui.leSnapshotName.text()).strip()

        if not BeamlineSnapshot().is_valid_name(snapshot_name):
            QtWidgets.QMessageBox.warning(self, "Error",
                                      "Snapshot name may consist of \
(alphanumeric_-=) characters only and cannot contain word 'general'!",
                                      QtWidgets.QMessageBox.Ok)
            return

        snapshot = BeamlineSnapshot(name=snapshot_name)

        for _, motor_tuple in enumerate(self.motor_tuples):                 # to be saved motors
            enabled_widget = motor_tuple[2]
            if enabled_widget.isChecked():
                step_widget = motor_tuple[3]

                dev_name, motor_name = motor_tuple[:2]
                position = self.beamline_hal.motor_position(f"{dev_name}/{motor_name}")
                snapshot.add_item(dev_name, motor_name, position, step_widget.value())

        if snapshot.empty():
            QtWidgets.QMessageBox.warning(self, "Error", "Select at least one motor!",
                                      QtWidgets.QMessageBox.Ok)
            return

        try:
            self.snapshot_manager.add_snapshot(snapshot)
            filename = f"{os.path.join(self.snapshot_manager.snapshots_dir(), snapshot_name)}.xml"
            snapshot.save(filename)
        except RuntimeError as err:
            # report_error() TODO
            QtWidgets.QMessageBox.warning(self, "Error", f"{str(err)}!",
                                      QtWidgets.QMessageBox.Ok)
        else:
            super().accept()

    # ----------------------------------------------------------------------
    @staticmethod
    def _next_color(devIdx):
        colors = [(220, 250, 220),
                  (250, 220, 220),
                  (250, 250, 220),
                  (220, 220, 250)]
        
        return colors[devIdx % len(colors)]


"""'About program' dialog
"""

from datetime import datetime
import os

from PyQt5 import QtWidgets

from beamline_console.gui.AboutDialog_ui import Ui_AboutDialog


# ----------------------------------------------------------------------
class AboutDialog(QtWidgets.QDialog):
    """
    """
    SOURCE_DIR = "src"
    DATETIME = "%Y-%m-%d %H:%M:%S"

    # ----------------------------------------------------------------------
    def __init__(self, parent):
        super().__init__(parent)

        self._ui = Ui_AboutDialog()
        self._ui.setupUi(self)

        # in case of multiple screens center on the primary one
        frame = self.frameGeometry()

        desktop = QtWidgets.QApplication.desktop()
        centerPoint = desktop.screenGeometry(desktop.primaryScreen()).center()

        frame.moveCenter(centerPoint)
        self.move(frame.topLeft())


"""Selection frame around device's graphicsitem, including title.
"""

from PyQt5 import QtWidgets

from beamline_console.constants import DEVICE_TITEL

# ----------------------------------------------------------------------
class TitleItem(QtWidgets.QGraphicsTextItem):

    # ----------------------------------------------------------------------
    def __init__(self, text, geometry):
        QtWidgets.QGraphicsTextItem.__init__(self, text)

        self._geometry = geometry
        self._width= int(self.textWidth())
        self.setPos(self._geometry[0], self._geometry[1])
        self.setFont(DEVICE_TITEL)

        #
        self.offsetX = 0
        self.offsetY = 0

    # ----------------------------------------------------------------------
    def updateOffsetX(self, offsetX):
        """
        """
        self.offsetX = offsetX
        self.setX(self._geometry[0] + self.offsetX)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'uis/ErrorDialog.ui'
#
# Created by: PyQt5 UI code generator 5.11.3
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_ErrorDialog(object):
    def setupUi(self, ErrorDialog):
        ErrorDialog.setObjectName("ErrorDialog")
        ErrorDialog.resize(950, 327)
        self.verticalLayout = QtWidgets.QVBoxLayout(ErrorDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.teDetails = QtWidgets.QTextEdit(ErrorDialog)
        self.teDetails.setEnabled(True)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.teDetails.sizePolicy().hasHeightForWidth())
        self.teDetails.setSizePolicy(sizePolicy)
        self.teDetails.setReadOnly(True)
        self.teDetails.setObjectName("teDetails")
        self.verticalLayout.addWidget(self.teDetails)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem)
        self.tbOk = QtWidgets.QToolButton(ErrorDialog)
        self.tbOk.setObjectName("tbOk")
        self.horizontalLayout_2.addWidget(self.tbOk)
        self.verticalLayout.addLayout(self.horizontalLayout_2)

        self.retranslateUi(ErrorDialog)
        self.tbOk.clicked.connect(ErrorDialog.accept)
        QtCore.QMetaObject.connectSlotsByName(ErrorDialog)

    def retranslateUi(self, ErrorDialog):
        _translate = QtCore.QCoreApplication.translate
        ErrorDialog.setWindowTitle(_translate("ErrorDialog", "Experimental Control Error"))
        self.tbOk.setText(_translate("ErrorDialog", "OK"))




# yury.matveev@desy.de

"""
Widget representing single motor.
"""

import logging
import sys
import time
import threading

from queue import Queue, Empty

from PyQt5 import QtWidgets, QtCore

from beamline_console.constants import APP_NAME
from beamline_console.devicecontrol.motorpropsdialog import MotorPropsDialog
from beamline_console.utils.input_dialog import InputDialog

from beamline_console.gui.MotorItemWidget_ui import Ui_MotorItemWidget

logger = logging.getLogger(APP_NAME)

CONTINUOUS_MOVE_DELAY = 0.5
CONTINUOUS_MOVE_REFRESH = 0.1


# ----------------------------------------------------------------------
class MotorItemWidget(QtWidgets.QWidget):
    """
    """

    # ----------------------------------------------------------------------
    def __init__(self, beamline_hal, device_handle, motor_handle, parent=None):
        super().__init__()

        self._beamline_hal = beamline_hal
        self._device_handle = device_handle
        self._motor_handle = motor_handle

        self._ui = Ui_MotorItemWidget()
        self._ui.setupUi(self)

        self._access_level = "user"

        self._ui.lb_name.setText(self.absolute_motor_name())
        if hasattr(self._motor_handle, "tango_server"):
            self._ui.lb_name.setToolTip(f"{self._motor_handle.tango_server}")

        self._ui.lb_name.double_clicked.connect(self._show_properties)

        self._ui.but_move_up.pressed.connect(lambda: self._move_pressed("up"))
        self._ui.but_move_up.released.connect(self._move_released)
        self._ui.but_move_up.grabGesture(QtCore.Qt.TapAndHoldGesture)
        self._ui.but_move_up.installEventFilter(EventFilter())

        self._ui.but_move_down.pressed.connect(lambda: self._move_pressed("down"))
        self._ui.but_move_down.released.connect(self._move_released)
        self._ui.but_move_down.grabGesture(QtCore.Qt.TapAndHoldGesture)
        self._ui.but_move_down.installEventFilter(EventFilter())

        self._ui.but_increase_step.clicked.connect(lambda: self._change_step("increase"))
        self._ui.but_decrease_step.clicked.connect(lambda: self._change_step("decrease"))
        self._ui.lb_pos.double_clicked.connect(self._move_motor_to)

        self._stop_continuous_move = threading.Event()
        self._moving_continuous = threading.Event()
        self._ignore_click = threading.Event()
        self._movement_worker = None
        self._last_clicks = Queue()

        self._move_tick = 0

        self._target_pos = 0

        self.refresh_gui()

    # ----------------------------------------------------------------------
    def access_level_changed(self, access_level):
        """
        """
        self._access_level = access_level

    # ----------------------------------------------------------------------
    def sync_pos(self, scale, mode="sync"):
        """
        Set "new position" QLineEdit to "current" QLineEdit.
        """
        DELAY = 0.2  # <<< TODO

        if mode.lower() == "sync":
            while self._motor_handle.state() == "moving":
                time.sleep(DELAY)

        if scale == "absolute":
            position = self._motor_handle.position()
        else:
            position = 0

        self._ui.lb_pos.setText(f"{position:.{self._motor_handle.decimals}f}")
        self._target_pos = position

    # ----------------------------------------------------------------------
    def _move_motor_to(self):
        """Move motor associated with the item
        """

        try:
            motor_name = self.absolute_motor_name()
            dlg = InputDialog(mode="position", motor_handle=self._motor_handle)
            if dlg.exec_():
                target_pos = dlg.new_position
                is_valid, status = self._motor_handle.is_move_valid(target_pos)

                if is_valid:
                    self._target_pos = target_pos
                    self._beamline_hal.move_motor(motor_name, target_pos, "absolute", "async")
                else:
                    logger.error(status)

        except Exception as err:
            logger.error(err, exc_info=sys.exc_info())
            raise

    # ----------------------------------------------------------------------
    def _change_step(self, direction):

        current_step = self._ui.sb_step.value()

        if direction == "increase":
            self._ui.sb_step.setValue(current_step*2)
        else:
            self._ui.sb_step.setValue(current_step/2)

    # ----------------------------------------------------------------------
    def _move_pressed(self, direction):
        self._stop_continuous_move.clear()
        self._moving_continuous.set()
        print("Put")
        self._last_clicks.put(1)

        if self._movement_worker is None or not self._movement_worker.is_alive():
            self._movement_worker = threading.Thread(target=self._do_move, args=[direction], daemon=True)
            self._movement_worker.start()

    # ----------------------------------------------------------------------
    def _move_released(self):
        self._stop_continuous_move.set()
        try:
            print("Get")
            self._last_clicks.get(block=False)
        except Empty:
            pass

    # ----------------------------------------------------------------------
    def _do_move(self, direction):
        time.sleep(CONTINUOUS_MOVE_DELAY)
        try:
            self._last_clicks.get(block=False)
            print("Got")
        except Empty:
            delta = abs(self._ui.sb_step.value())
            if direction == "down":
                delta *= -1
            try:
                position = self._motor_handle.position()
                self._target_pos = position + delta
                self._beamline_hal.move_motor(self.absolute_motor_name(), self._target_pos, "absolute")

            except Exception as err:
                logger.error(err, exc_info=sys.exc_info())
        else:
            if not self._stop_continuous_move.is_set():
                self._ignore_click.set()
                logger.info(f"Start continuous move, direction {direction}")
                try:
                    if direction == "up":
                        self._motor_handle.move_to_high_limit()
                    else:
                        self._motor_handle.move_to_low_limit()
                except Exception as err:
                    logger.error(f'Cannot move to limit: {repr(err)}')

                while not self._stop_continuous_move.is_set():
                    time.sleep(CONTINUOUS_MOVE_REFRESH)

                logger.info(f"Stop continuous move")
                self._motor_handle.stop()
                self._target_pos = self._motor_handle.position()

            self._moving_continuous.clear()

    # ----------------------------------------------------------------------
    def refresh_gui(self):
        """Called periodically by the DeviceWidget.
        """

        decimals = self._motor_handle.decimals
        position = self._motor_handle.position()
        try:
            self._ui.lb_pos.setText(f"{position:.{decimals}f}")

            motor_state, limits_state = self.get_status()
            self._colorize_diodes(motor_state, limits_state)

            self._ui.le_low_limit.setText(f"{self._motor_handle.get_low_limit():.{decimals}f}")
            self._ui.le_high_limit.setText(f"{self._motor_handle.get_high_limit():.{decimals}f}")

            low_lim = self._motor_handle.get_low_limit()
            slider_range = self._motor_handle.get_high_limit() - low_lim
            self._ui.sl_position.setMaximum(slider_range)
            self._ui.sl_position.setSingleStep(slider_range/100)

            if not self._moving_continuous.is_set():
                self._ui.but_move_up.setEnabled(motor_state != "moving")
                self._ui.but_move_down.setEnabled(motor_state != "moving")

            if self._target_pos < position:
                self._ui.sl_position.setLow(self._target_pos-low_lim)
                self._ui.sl_position.setHigh(position-low_lim)
            else:
                self._ui.sl_position.setLow(position-low_lim)
                self._ui.sl_position.setHigh(self._target_pos-low_lim)

            self._ui.sb_step.setDecimals(decimals)

        except Exception:  # PyTango.DevFailed:
            self._colorize_diodes("fault", '')

    # ----------------------------------------------------------------------
    def get_status(self):

        return self._motor_handle.state(), self._motor_handle.check_limits()

    # ----------------------------------------------------------------------
    def _colorize_diodes(self, motor_state, limits_state):
        """Update the "diodes" signaling associated motor's state and update status label
        """
        error_style = "QToolButton {background-color: rgb(242, 101, 101);}"

        lim_green_style = "QToolButton {background-color: rgb(103, 214, 98);}"
        lim_red_style = "QToolButton {background-color: rgb(242, 101, 101);}"

        low_style = lim_green_style
        low_status = 'Limit is OK'

        high_style = lim_green_style
        high_status = 'Limit is OK'

        if motor_state == "moving":
            self._move_tick += 1  # make diode blinking
            if self._move_tick % 3 == 0:
                st_style = "QToolButton {background-color: rgb(103, 98, 214);}"
            elif self._move_tick % 3 == 1:
                st_style = "QToolButton {background-color: rgb(186, 48, 176);}"
            else:
                st_style = "QToolButton {background-color: rgb(191, 197, 42);}"

            st_tip = "Motor moving"

        # elif motor_state != "on":
        #     self._move_tick = 0
        #     st_style = error_style
        #     st_tip = "Motor OFF"
        else:
            st_style = lim_green_style
            st_tip = "Motor ON"

        if len(limits_state) > 1:
            st_style = error_style
            st_tip =  "Both Limits are touched"

            low_style = high_style = lim_red_style
            low_status = high_status = 'Limit touched'
        elif len(limits_state):
            st_style = error_style

            if limits_state[0] == 'low':
                st_tip = "Low limit are touched"
                low_style = lim_red_style
                low_status = 'Limit touched'
            else:
                st_tip = "High limit are touched"
                high_style = lim_red_style
                high_status = 'Limit touched'

        self._ui.tb_status.setStyleSheet(st_style)
        self._ui.tb_status.setToolTip(st_tip)

        self._ui.tb_low_limit.setStyleSheet(low_style)
        self._ui.tb_low_limit.setToolTip(low_status)

        self._ui.tb_high_limit.setStyleSheet(high_style)
        self._ui.tb_high_limit.setToolTip(high_status)

    # ----------------------------------------------------------------------
    def absolute_motor_name(self):
        """
        Returns:
            (str) name of the motor associated with the widget
        """
        return f"{self._device_handle.name}/{self._motor_handle.name}"

    # ----------------------------------------------------------------------
    def _show_properties(self):
        """
        """
        if self._access_level == "superuser":
            dialog = MotorPropsDialog(self._motor_handle, self)
            dialog.show()


class EventFilter(QtCore.QObject):
    def eventFilter(self, parent, event):
        print(f"Got event: {event} from {parent}")


# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'uis/MotorItemWidget.ui'
#
# Created by: PyQt5 UI code generator 5.11.3
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_MotorItemWidget(object):
    def setupUi(self, MotorItemWidget):
        MotorItemWidget.setObjectName("MotorItemWidget")
        MotorItemWidget.resize(700, 173)
        MotorItemWidget.setMinimumSize(QtCore.QSize(700, 0))
        MotorItemWidget.setMaximumSize(QtCore.QSize(16777215, 16777215))
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(MotorItemWidget)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem)
        self.tb_status = QtWidgets.QToolButton(MotorItemWidget)
        self.tb_status.setEnabled(False)
        self.tb_status.setMaximumSize(QtCore.QSize(18, 14))
        self.tb_status.setStyleSheet("background-color: rgb(85, 170, 0);")
        self.tb_status.setText("")
        self.tb_status.setObjectName("tb_status")
        self.horizontalLayout_2.addWidget(self.tb_status)
        self.lb_name = ClickableLabel(MotorItemWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lb_name.sizePolicy().hasHeightForWidth())
        self.lb_name.setSizePolicy(sizePolicy)
        self.lb_name.setMinimumSize(QtCore.QSize(160, 0))
        self.lb_name.setMaximumSize(QtCore.QSize(1000, 16777215))
        font = QtGui.QFont()
        font.setPointSize(16)
        font.setBold(True)
        font.setWeight(75)
        self.lb_name.setFont(font)
        self.lb_name.setObjectName("lb_name")
        self.horizontalLayout_2.addWidget(self.lb_name)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem1)
        self.verticalLayout_2.addLayout(self.horizontalLayout_2)
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.but_move_down = QtWidgets.QToolButton(MotorItemWidget)
        self.but_move_down.setMinimumSize(QtCore.QSize(90, 120))
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/left_arrow.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.but_move_down.setIcon(icon)
        self.but_move_down.setIconSize(QtCore.QSize(65, 65))
        self.but_move_down.setObjectName("but_move_down")
        self.horizontalLayout_5.addWidget(self.but_move_down)
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_4.addItem(spacerItem2)
        self.label = QtWidgets.QLabel(MotorItemWidget)
        font = QtGui.QFont()
        font.setPointSize(14)
        self.label.setFont(font)
        self.label.setObjectName("label")
        self.horizontalLayout_4.addWidget(self.label)
        self.but_decrease_step = QtWidgets.QToolButton(MotorItemWidget)
        self.but_decrease_step.setMinimumSize(QtCore.QSize(75, 50))
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap(":/icons/minus.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.but_decrease_step.setIcon(icon1)
        self.but_decrease_step.setIconSize(QtCore.QSize(45, 45))
        self.but_decrease_step.setObjectName("but_decrease_step")
        self.horizontalLayout_4.addWidget(self.but_decrease_step)
        self.sb_step = QtWidgets.QDoubleSpinBox(MotorItemWidget)
        self.sb_step.setMaximumSize(QtCore.QSize(110, 16777215))
        font = QtGui.QFont()
        font.setPointSize(14)
        self.sb_step.setFont(font)
        self.sb_step.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.sb_step.setDecimals(3)
        self.sb_step.setMaximum(999.0)
        self.sb_step.setSingleStep(0.1)
        self.sb_step.setProperty("value", 0.1)
        self.sb_step.setObjectName("sb_step")
        self.horizontalLayout_4.addWidget(self.sb_step)
        self.but_increase_step = QtWidgets.QToolButton(MotorItemWidget)
        self.but_increase_step.setMinimumSize(QtCore.QSize(75, 50))
        icon2 = QtGui.QIcon()
        icon2.addPixmap(QtGui.QPixmap(":/icons/plus.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.but_increase_step.setIcon(icon2)
        self.but_increase_step.setIconSize(QtCore.QSize(45, 45))
        self.but_increase_step.setObjectName("but_increase_step")
        self.horizontalLayout_4.addWidget(self.but_increase_step)
        spacerItem3 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_4.addItem(spacerItem3)
        self.verticalLayout.addLayout(self.horizontalLayout_4)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.le_low_limit = QtWidgets.QLineEdit(MotorItemWidget)
        self.le_low_limit.setEnabled(False)
        self.le_low_limit.setMaximumSize(QtCore.QSize(110, 16777215))
        self.le_low_limit.setStyleSheet("background-color: rgb(218, 221, 224);")
        self.le_low_limit.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.le_low_limit.setReadOnly(True)
        self.le_low_limit.setObjectName("le_low_limit")
        self.horizontalLayout.addWidget(self.le_low_limit)
        self.tb_low_limit = QtWidgets.QToolButton(MotorItemWidget)
        self.tb_low_limit.setEnabled(False)
        self.tb_low_limit.setMaximumSize(QtCore.QSize(18, 14))
        self.tb_low_limit.setStyleSheet("background-color: rgb(85, 170, 0);")
        self.tb_low_limit.setText("")
        self.tb_low_limit.setObjectName("tb_low_limit")
        self.horizontalLayout.addWidget(self.tb_low_limit)
        self.sl_position = RangeSlider(MotorItemWidget)
        self.sl_position.setEnabled(True)
        self.sl_position.setOrientation(QtCore.Qt.Horizontal)
        self.sl_position.setObjectName("sl_position")
        self.horizontalLayout.addWidget(self.sl_position)
        self.tb_high_limit = QtWidgets.QToolButton(MotorItemWidget)
        self.tb_high_limit.setEnabled(False)
        self.tb_high_limit.setMaximumSize(QtCore.QSize(18, 14))
        self.tb_high_limit.setStyleSheet("background-color: rgb(85, 170, 0);")
        self.tb_high_limit.setText("")
        self.tb_high_limit.setObjectName("tb_high_limit")
        self.horizontalLayout.addWidget(self.tb_high_limit)
        self.le_high_limit = QtWidgets.QLineEdit(MotorItemWidget)
        self.le_high_limit.setEnabled(False)
        self.le_high_limit.setMaximumSize(QtCore.QSize(110, 16777215))
        self.le_high_limit.setStyleSheet("background-color: rgb(218, 221, 224);")
        self.le_high_limit.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.le_high_limit.setReadOnly(True)
        self.le_high_limit.setObjectName("le_high_limit")
        self.horizontalLayout.addWidget(self.le_high_limit)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        spacerItem4 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem4)
        self.label_2 = QtWidgets.QLabel(MotorItemWidget)
        font = QtGui.QFont()
        font.setPointSize(14)
        self.label_2.setFont(font)
        self.label_2.setObjectName("label_2")
        self.horizontalLayout_3.addWidget(self.label_2)
        self.lb_pos = ClickableLabel(MotorItemWidget)
        font = QtGui.QFont()
        font.setPointSize(14)
        self.lb_pos.setFont(font)
        self.lb_pos.setObjectName("lb_pos")
        self.horizontalLayout_3.addWidget(self.lb_pos)
        spacerItem5 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem5)
        self.verticalLayout.addLayout(self.horizontalLayout_3)
        self.horizontalLayout_5.addLayout(self.verticalLayout)
        self.but_move_up = QtWidgets.QToolButton(MotorItemWidget)
        self.but_move_up.setMinimumSize(QtCore.QSize(90, 120))
        icon3 = QtGui.QIcon()
        icon3.addPixmap(QtGui.QPixmap(":/icons/rigth_arrow.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.but_move_up.setIcon(icon3)
        self.but_move_up.setIconSize(QtCore.QSize(65, 65))
        self.but_move_up.setObjectName("but_move_up")
        self.horizontalLayout_5.addWidget(self.but_move_up)
        self.verticalLayout_2.addLayout(self.horizontalLayout_5)

        self.retranslateUi(MotorItemWidget)
        QtCore.QMetaObject.connectSlotsByName(MotorItemWidget)

    def retranslateUi(self, MotorItemWidget):
        _translate = QtCore.QCoreApplication.translate
        MotorItemWidget.setWindowTitle(_translate("MotorItemWidget", "Form"))
        self.tb_status.setToolTip(_translate("MotorItemWidget", "Motor ON"))
        self.lb_name.setText(_translate("MotorItemWidget", "Motor name:"))
        self.but_move_down.setToolTip(_translate("MotorItemWidget", "Move by (-) Delta"))
        self.but_move_down.setText(_translate("MotorItemWidget", "<"))
        self.label.setText(_translate("MotorItemWidget", "step:"))
        self.but_decrease_step.setToolTip(_translate("MotorItemWidget", "Move by (-) Delta"))
        self.but_decrease_step.setText(_translate("MotorItemWidget", "<"))
        self.sb_step.setToolTip(_translate("MotorItemWidget", "Delta"))
        self.but_increase_step.setToolTip(_translate("MotorItemWidget", "Move by (+) Delta"))
        self.but_increase_step.setText(_translate("MotorItemWidget", ">"))
        self.le_low_limit.setToolTip(_translate("MotorItemWidget", "Low Limit "))
        self.tb_low_limit.setToolTip(_translate("MotorItemWidget", "Low limit OK"))
        self.tb_high_limit.setToolTip(_translate("MotorItemWidget", "High limit OK"))
        self.le_high_limit.setToolTip(_translate("MotorItemWidget", "High Limit"))
        self.label_2.setText(_translate("MotorItemWidget", "position: "))
        self.lb_pos.setText(_translate("MotorItemWidget", "0.001"))
        self.but_move_up.setToolTip(_translate("MotorItemWidget", "Move by (+) Delta"))
        self.but_move_up.setText(_translate("MotorItemWidget", ">"))

from beamline_console.utils.clickablelabel import ClickableLabel
from beamline_console.utils.range_slider import RangeSlider
import beamline_console.gui.icons_rc



"""Logger propagating log records to the LoggingWidget.
"""

from logging import StreamHandler
from PyQt5 import QtCore


# ----------------------------------------------------------------------
class GuiLogger(QtCore.QObject, StreamHandler):

    emit_record = QtCore.pyqtSignal(int, str)

    # ----------------------------------------------------------------------
    def __init__(self):
        QtCore.QObject.__init__(self)
        StreamHandler.__init__(self)

    # ----------------------------------------------------------------------
    def emit(self, record):
        self.emit_record.emit(int(record.levelno), f'{record.asctime.split(",")[0]} {record.levelname} {record.msg}')



# yury.matveev@desy.de

from PyQt5 import QtGui, QtCore

from beamline_console.constants import DEBUG_COLOR, INFO_COLOR, WARNING_COLOR
from beamline_console.constants import ERROR_COLOR, DATE_COLOR


# ----------------------------------------------------------------------
class HighlightLogs(QtGui.QSyntaxHighlighter):

    # ----------------------------------------------------------------------
    def __init__(self, document):
        super().__init__(document)

        self.rules = []             # (pattern, style) pairs

        self._add_rule("\\bDEBUG\\b", DEBUG_COLOR)
        self._add_rule("\\bINFO\\b", INFO_COLOR)
        self._add_rule("\\bWARNING\\b", WARNING_COLOR)
        self._add_rule("\\bERROR\\b", ERROR_COLOR)

        self._add_rule("\\b[0-9]{4}-[0-9]{2}-[0-9]{2} [0-9]{2}:[0-9]{2}:[0-9]{2}\\b", DATE_COLOR)

    # ----------------------------------------------------------------------
    def _add_rule(self, token, color):
        """
        """
        style = QtGui.QTextCharFormat()
        style.setForeground(color)
        
        self.rules.append((QtCore.QRegExp(token), style))

    # ----------------------------------------------------------------------
    def highlightBlock(self, text):
        """Highlight block of text according to defined rules
        """
        for pattern, style in self.rules:
            expression = QtCore.QRegExp(pattern)
            index = expression.indexIn(text, 0)

            while index >= 0:
                length = expression.matchedLength()
                self.setFormat(index, length, style)
                index = expression.indexIn(text, index + length)

        self.setCurrentBlockState(0)




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'uis/InputDialog.ui'
#
# Created by: PyQt5 UI code generator 5.11.3
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_InputDialog(object):
    def setupUi(self, InputDialog):
        InputDialog.setObjectName("InputDialog")
        InputDialog.resize(700, 700)
        InputDialog.setMinimumSize(QtCore.QSize(700, 700))
        InputDialog.setMaximumSize(QtCore.QSize(700, 700))
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(InputDialog)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.lb_titel = QtWidgets.QLabel(InputDialog)
        font = QtGui.QFont()
        font.setPointSize(20)
        font.setBold(True)
        font.setWeight(75)
        self.lb_titel.setFont(font)
        self.lb_titel.setAlignment(QtCore.Qt.AlignCenter)
        self.lb_titel.setObjectName("lb_titel")
        self.verticalLayout.addWidget(self.lb_titel)
        self.lb_number = QtWidgets.QLabel(InputDialog)
        font = QtGui.QFont()
        font.setPointSize(20)
        font.setBold(True)
        font.setWeight(75)
        self.lb_number.setFont(font)
        self.lb_number.setAlignment(QtCore.Qt.AlignCenter)
        self.lb_number.setObjectName("lb_number")
        self.verticalLayout.addWidget(self.lb_number)
        self.horizontalLayout.addLayout(self.verticalLayout)
        self.verticalLayout_2.addLayout(self.horizontalLayout)
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setSpacing(20)
        self.gridLayout.setObjectName("gridLayout")
        self.but_1 = QtWidgets.QPushButton(InputDialog)
        self.but_1.setMinimumSize(QtCore.QSize(100, 100))
        self.but_1.setMaximumSize(QtCore.QSize(100, 100))
        font = QtGui.QFont()
        font.setPointSize(18)
        font.setBold(True)
        font.setWeight(75)
        self.but_1.setFont(font)
        self.but_1.setObjectName("but_1")
        self.gridLayout.addWidget(self.but_1, 0, 0, 1, 1)
        self.but_2 = QtWidgets.QPushButton(InputDialog)
        self.but_2.setMinimumSize(QtCore.QSize(100, 100))
        self.but_2.setMaximumSize(QtCore.QSize(100, 100))
        font = QtGui.QFont()
        font.setPointSize(18)
        font.setBold(True)
        font.setWeight(75)
        self.but_2.setFont(font)
        self.but_2.setObjectName("but_2")
        self.gridLayout.addWidget(self.but_2, 0, 1, 1, 1)
        self.but_3 = QtWidgets.QPushButton(InputDialog)
        self.but_3.setMinimumSize(QtCore.QSize(100, 100))
        self.but_3.setMaximumSize(QtCore.QSize(100, 100))
        font = QtGui.QFont()
        font.setPointSize(18)
        font.setBold(True)
        font.setWeight(75)
        self.but_3.setFont(font)
        self.but_3.setObjectName("but_3")
        self.gridLayout.addWidget(self.but_3, 0, 2, 1, 1)
        self.but_4 = QtWidgets.QPushButton(InputDialog)
        self.but_4.setMinimumSize(QtCore.QSize(100, 100))
        self.but_4.setMaximumSize(QtCore.QSize(100, 100))
        font = QtGui.QFont()
        font.setPointSize(18)
        font.setBold(True)
        font.setWeight(75)
        self.but_4.setFont(font)
        self.but_4.setObjectName("but_4")
        self.gridLayout.addWidget(self.but_4, 1, 0, 1, 1)
        self.but_5 = QtWidgets.QPushButton(InputDialog)
        self.but_5.setMinimumSize(QtCore.QSize(100, 100))
        self.but_5.setMaximumSize(QtCore.QSize(100, 100))
        font = QtGui.QFont()
        font.setPointSize(18)
        font.setBold(True)
        font.setWeight(75)
        self.but_5.setFont(font)
        self.but_5.setObjectName("but_5")
        self.gridLayout.addWidget(self.but_5, 1, 1, 1, 1)
        self.but_6 = QtWidgets.QPushButton(InputDialog)
        self.but_6.setMinimumSize(QtCore.QSize(100, 100))
        self.but_6.setMaximumSize(QtCore.QSize(100, 100))
        font = QtGui.QFont()
        font.setPointSize(18)
        font.setBold(True)
        font.setWeight(75)
        self.but_6.setFont(font)
        self.but_6.setObjectName("but_6")
        self.gridLayout.addWidget(self.but_6, 1, 2, 1, 1)
        self.but_7 = QtWidgets.QPushButton(InputDialog)
        self.but_7.setMinimumSize(QtCore.QSize(100, 100))
        self.but_7.setMaximumSize(QtCore.QSize(100, 100))
        font = QtGui.QFont()
        font.setPointSize(18)
        font.setBold(True)
        font.setWeight(75)
        self.but_7.setFont(font)
        self.but_7.setObjectName("but_7")
        self.gridLayout.addWidget(self.but_7, 2, 0, 1, 1)
        self.but_8 = QtWidgets.QPushButton(InputDialog)
        self.but_8.setMinimumSize(QtCore.QSize(100, 100))
        self.but_8.setMaximumSize(QtCore.QSize(100, 100))
        font = QtGui.QFont()
        font.setPointSize(18)
        font.setBold(True)
        font.setWeight(75)
        self.but_8.setFont(font)
        self.but_8.setObjectName("but_8")
        self.gridLayout.addWidget(self.but_8, 2, 1, 1, 1)
        self.but_9 = QtWidgets.QPushButton(InputDialog)
        self.but_9.setMinimumSize(QtCore.QSize(100, 100))
        self.but_9.setMaximumSize(QtCore.QSize(100, 100))
        font = QtGui.QFont()
        font.setPointSize(18)
        font.setBold(True)
        font.setWeight(75)
        self.but_9.setFont(font)
        self.but_9.setObjectName("but_9")
        self.gridLayout.addWidget(self.but_9, 2, 2, 1, 1)
        self.but_0 = QtWidgets.QPushButton(InputDialog)
        self.but_0.setMinimumSize(QtCore.QSize(100, 100))
        self.but_0.setMaximumSize(QtCore.QSize(100, 100))
        font = QtGui.QFont()
        font.setPointSize(18)
        font.setBold(True)
        font.setWeight(75)
        self.but_0.setFont(font)
        self.but_0.setObjectName("but_0")
        self.gridLayout.addWidget(self.but_0, 3, 1, 1, 1)
        self.but_punkt = QtWidgets.QPushButton(InputDialog)
        self.but_punkt.setMinimumSize(QtCore.QSize(100, 100))
        self.but_punkt.setMaximumSize(QtCore.QSize(100, 100))
        font = QtGui.QFont()
        font.setPointSize(18)
        font.setBold(True)
        font.setWeight(75)
        self.but_punkt.setFont(font)
        self.but_punkt.setObjectName("but_punkt")
        self.gridLayout.addWidget(self.but_punkt, 3, 0, 1, 1)
        self.but_sign = QtWidgets.QPushButton(InputDialog)
        self.but_sign.setMinimumSize(QtCore.QSize(100, 100))
        self.but_sign.setMaximumSize(QtCore.QSize(100, 100))
        font = QtGui.QFont()
        font.setPointSize(18)
        font.setBold(True)
        font.setWeight(75)
        self.but_sign.setFont(font)
        self.but_sign.setObjectName("but_sign")
        self.gridLayout.addWidget(self.but_sign, 3, 2, 1, 1)
        self.verticalLayout_2.addLayout(self.gridLayout)
        self.but_box = QtWidgets.QDialogButtonBox(InputDialog)
        font = QtGui.QFont()
        font.setPointSize(20)
        self.but_box.setFont(font)
        self.but_box.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok|QtWidgets.QDialogButtonBox.Reset)
        self.but_box.setCenterButtons(False)
        self.but_box.setObjectName("but_box")
        self.verticalLayout_2.addWidget(self.but_box)

        self.retranslateUi(InputDialog)
        QtCore.QMetaObject.connectSlotsByName(InputDialog)

    def retranslateUi(self, InputDialog):
        _translate = QtCore.QCoreApplication.translate
        InputDialog.setWindowTitle(_translate("InputDialog", "Experiment Control"))
        self.lb_titel.setText(_translate("InputDialog", "Enter superuser pin"))
        self.lb_number.setText(_translate("InputDialog", "*****"))
        self.but_1.setText(_translate("InputDialog", "1"))
        self.but_2.setText(_translate("InputDialog", "2"))
        self.but_3.setText(_translate("InputDialog", "3"))
        self.but_4.setText(_translate("InputDialog", "4"))
        self.but_5.setText(_translate("InputDialog", "5"))
        self.but_6.setText(_translate("InputDialog", "6"))
        self.but_7.setText(_translate("InputDialog", "7"))
        self.but_8.setText(_translate("InputDialog", "8"))
        self.but_9.setText(_translate("InputDialog", "9"))
        self.but_0.setText(_translate("InputDialog", "0"))
        self.but_punkt.setText(_translate("InputDialog", ","))
        self.but_sign.setText(_translate("InputDialog", "+/-"))




# yury.matveev@desy.de
"""
Class used for management of beamline state snapshots.
"""

import os
import shutil

from PyQt5 import QtCore

from beamline_console.logger.logit import Logit
from beamline_console.snapshot.beamsnapshot import BeamlineSnapshot

# ----------------------------------------------------------------------
class SnapshotManager(QtCore.QObject, Logit):
    """
    """
    changed = QtCore.pyqtSignal()

    # ---------------------------------------------------------------------- 
    def __init__(self, beamline_hal, snapshots_dir=""):
        """
        """
        QtCore.QObject.__init__(self)
        Logit.__init__(self)

        self._snapshots_dir = ""
        self._trash_dir = ""

        self._beamline_hal = beamline_hal
        self.set_snapshots_dir(snapshots_dir)

        self.beamline_snapshots = []

        self._current_snapshot = None

    # ----------------------------------------------------------------------
    def add_snapshot(self, snapshot):
        """
        """
        if snapshot.name in [s.name for s in self.beamline_snapshots]: 
            raise RuntimeError(f"Snapshot {snapshot.name} already exists")

        self.beamline_snapshots.append(snapshot)
    
    # ----------------------------------------------------------------------
    def update_snapshot(self, snapshot_name):
        """Update snapshot to current position.
        """
        self.logger.info(f"Snapshot {snapshot_name} updated")

        for snapshot_idx, snapshot in enumerate(self.beamline_snapshots):
            if snapshot.name == snapshot_name:
                os.remove(snapshot.filename)
                snapshot.update_position()
                snapshot.save(snapshot.filename)

    # ----------------------------------------------------------------------
    def delete_snapshot(self, snapshot_name):
        """Move snapshot to the "trash directory".
        """
        self.logger.debug(f"Removing snapshot {snapshot_name}")

        for snapshot_idx, snapshot in enumerate(self.beamline_snapshots):
            if snapshot.name == snapshot_name:
                dest_fname = os.path.join(self._trash_dir, os.path.basename(snapshot.filename))
                if os.path.exists(dest_fname):
                    os.remove(dest_fname)

                shutil.move(snapshot.filename, self._trash_dir)
        
                del self.beamline_snapshots[snapshot_idx]
                break

    # ---------------------------------------------------------------------- 
    def load(self, snapshots_dir):
        """Reload snapshots from the snapshots directory.
        """
        self.set_snapshots_dir(snapshots_dir)
        self._reload_all(sorted(os.listdir(snapshots_dir)))

    # ----------------------------------------------------------------------
    def _reload_all(self, file_list):
        """
        """
        self.beamline_snapshots = []
    
        for filename in file_list:
            full_name = os.path.join(self._snapshots_dir, filename)
            if os.path.isdir(full_name):
                continue

            snap_name, ext = os.path.splitext(filename)
            if ext.lower() == ".xml" and snap_name:
                snapshot = BeamlineSnapshot(self._beamline_hal)
                if snapshot.load(full_name):
                    self.add_snapshot(snapshot)
 
    # ----------------------------------------------------------------------
    def empty(self):
        """
        Returns:
            (bool)
        """
        return len(self.beamline_snapshots) < 1  

    # ----------------------------------------------------------------------
    def set_snapshots_dir(self, snapshots_dir):
        """Set directory used for beamline snapshots and update "trash" accordingly.
        """
        self._snapshots_dir = snapshots_dir
        self._trash_dir = os.path.join(snapshots_dir, "_trash")
    
        if not os.path.exists(self._trash_dir):
            os.mkdir(self._trash_dir)

    # ----------------------------------------------------------------------
    def snapshots_dir(self):
        """
        Returns:
            (str)
        """
        return self._snapshots_dir

    # ----------------------------------------------------------------------
    def __len__(self):
        """
        """
        return len(self.beamline_snapshots)


    # NOTE: here, SnapshotManager acts as proxy for the currently being executed snapshot!
    # ----------------------------------------------------------------------
    def state(self):
        """
        Returns:
            (str) state of the currently being executed snapshot
        """
        if not self._current_snapshot:
            return "idle"
       
        return self._current_snapshot.state

    # ----------------------------------------------------------------------
    def pause(self):
        """Pause currently executed snapshot - works in async mode only!
        """
        if self._current_snapshot:
            self._current_snapshot.pause()
     
    # ----------------------------------------------------------------------
    def resume(self):
        """Pause currently executed snapshot - works in async mode only!
        """
        if self._current_snapshot:
            self._current_snapshot.resume()

    # ----------------------------------------------------------------------
    def abort(self):
        """Pause currently executed snapshot - works in async mode only!
        """
        if self._current_snapshot:
            self._current_snapshot.abort()



# yury.matveev@desy.de

import logging

from beamline_console.constants import APP_NAME

# ----------------------------------------------------------------------
class Logit:
    def __init__(self):
        self.logger = logging.getLogger(APP_NAME)



"""Selection frame around device's graphicsitem, including title.
"""

from PyQt5 import QtWidgets, QtGui, QtCore

from beamline_console.logger.logit import Logit

# ----------------------------------------------------------------------
class FrameItem(QtWidgets.QGraphicsRectItem, Logit):

        # adjust to your needs and taste
    INACTIVE_PEN = QtGui.QPen(QtGui.QBrush(QtGui.QColor(170, 170, 170)), 2,
                              QtCore.Qt.DotLine, QtCore.Qt.SquareCap,
                              QtCore.Qt.BevelJoin)
    ACTIVE_PEN = QtGui.QPen(QtGui.QBrush(QtGui.QColor(30, 144, 255, 120)), 2,
                            QtCore.Qt.SolidLine, QtCore.Qt.SquareCap,
                            QtCore.Qt.BevelJoin)
    
    INACTIVE_BRUSH = QtGui.QBrush(QtGui.QColor(30, 144, 255, 10))
    ACTIVE_BRUSH = QtGui.QBrush(QtGui.QColor(30, 144, 255, 50))

    # ----------------------------------------------------------------------
    def __init__(self, geometry):
        QtWidgets.QGraphicsRectItem.__init__(self, *geometry)
        Logit.__init__(self)
       
        self._initialGeom = self.rect()
        
        self.setPen(self.INACTIVE_PEN)
        self.setBrush(self.ACTIVE_BRUSH)

            # 
        self.offsetX = 0
        self.offsetY = 0
        
    # ----------------------------------------------------------------------
    def setActive(self, isActive):
        """
        """
        self.setPen(self.ACTIVE_PEN if isActive else self.INACTIVE_PEN)
        self.setBrush(self.ACTIVE_BRUSH if isActive else self.INACTIVE_BRUSH)
       
    # ----------------------------------------------------------------------
    def updateOffsetX(self, offsetX):
        """
        """
        self.offsetX = offsetX
        
        rect = self.rect()
        rect.setLeft(self._initialGeom.left() + self.offsetX)
        rect.setRight(self._initialGeom.right() + self.offsetX)
        
        self.setRect(rect)
        
